﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.PlayerActions;
using Buildcraft.Core.Components;

using Buildcraft.Core;
using Buildcraft.ScenarioBuilder;

namespace Buildcraft.AI
{
    /// <summary>
    /// Class that handles player data
    /// </summary>
    public class AIPlayer
    {
        /// <summary>
        /// The player this AI module is working on
        /// </summary>
        private Player Player;

        private World World;

        // Player Actions. These are the only way that the AI can manipulate the world.
        private PlayerActions _playerActions;

        //AI View
        private List<AINode> AINodes = new List<AINode>();
        private List<Worker> TempWorkerList = new List<Worker>();

        //TODO: Current Objective State (e.g. current target node, current focused resource, etc)

        /// <summary>
        /// Run a recalculate every 100 ticks
        /// </summary>
        private Counter RecalcCounter = new Counter(50);
        private IBuildcraftContainer Container;


        public AIPlayer(Player player, IBuildcraftContainer container)
        {
            if (player == null)
                throw new ArgumentNullException("player");

            if (container == null)
                throw new ArgumentNullException("container");

            Player = player;
            Container = container;
            World = container.RetrieveInstance<World>();

            //Create Player Actions
            _playerActions = new PlayerActions(player, container);

            BuildAiWorld(World);
        }


        /// <summary>
        /// Builds the AI Interpretation of the game world.
        /// </summary>
        /// <param name="world">The game world to build a representation for.</param>
        private void BuildAiWorld(World world)
        {
            //Build up AI World Representation

            //Create AI Nodes
            foreach (var node in world.Nodes)
            {
                AINode aiNode = new AINode();
                aiNode.GameNode = node;

                AINodes.Add(aiNode);
            }

            //Create Joins for AI Nodes
            foreach (var aiNode in AINodes)
            {
                var gameNode = aiNode.GameNode;
                
                //Lookup adjacent nodes
                foreach (var join in gameNode.Joins)
                {
                    var targetGameNode = join.ToNode;
                    var targetAiNode = AINodes.Where(n => n.GameNode == targetGameNode).Single();

                    aiNode.Joins.Add(new AIJoin(aiNode, targetAiNode));
                }
            }
        }

        public void Tick(int ticks)
        {
            //Tick Recalc, when it fires perform recalculate.
            if (RecalcCounter.Tick(ticks))
                Recalc();
        }

        private void Recalc()
        {
            RecalculateWorld();
            RecalculateGoal();
        }


        private void RecalculateWorld()
        {
            //TODO: Weight node based on resource calculations
            //Determine resources demand
            //Determine which Resources we have a shortage of (Relative to demand)
            //Determine What resources we have a surplus of (Relative to demand)
            //Arrange workers on improvements

            //This would be quite expensive and need to be done preiodically.
            foreach (var aiNode in AINodes)
            {
                var gameNode = aiNode.GameNode;

                aiNode.Friendly = gameNode.Owner == Player;
                aiNode.Value = EvaluateNode(gameNode);
            }
        }

        /// <summary>
        /// Determine the value for a given node.
        /// </summary>
        /// <param name="gameNode"></param>
        /// <returns></returns>
        private int EvaluateNode(Node gameNode)
        {
            //Get Core Heuristic Evaluator
            float multiplier = CalculateNodeMultiplier(gameNode);
            float value = CalculateNodeValue(gameNode);

            return (int)(value * multiplier);
        }

        /// <summary> 
        /// Get the multiplier for a given node. The multiplier is the adjustment made to a node based on player .
        /// </summary>
        /// <param name="gameNode"></param>
        /// <returns></returns>
        private float CalculateNodeMultiplier(Node gameNode)
        {
            float multiplier = 1f;

            //Add weighting, for already owned nodes.
            if (gameNode.Owner == Player)
                multiplier += 0.25f;

            //Add Weighting For Adjacent Friendly nodes
            foreach (var join in gameNode.Joins)
            {
                var node = join.ToNode;

                if (node.Owner == Player)
                    multiplier += 0.1f;
            }

            return multiplier;
        }

        /// <summary>
        /// Determine the base value of a node, without any player specific context.
        /// </summary>
        /// <param name="gameNode"></param>
        /// <returns></returns>
        private float CalculateNodeValue(Node gameNode)
        {
            int value = 10;

            //Calculate Bonus for improvements
            if (gameNode.Improvement != null)
            {
                if (gameNode.Improvement.TypeInfo.ImprovementTypeCode == ImprovementTypeCodes.City)
                    value += 10;
                else
                    value += 5;
            }

            //TODO: Update bonus based on Defence bonus

            //TODO: Update bonus based on resource / richness

            return value;
        }


        private void RecalculateGoal()
        {
            //Determine Priority of tasks

            //Determine priority of production

            //Run Tasks
            RunTask(NodeTaskTypeCodes.AttackNode);
            RunTask(NodeTaskTypeCodes.WorkImprovement);
            RunTask(NodeTaskTypeCodes.DefendTerritory);
        }


        private void RunTask(string taskType)
        {
            //Get task type info, and determine best target
            var taskTypeInfo = Container.RetrieveAllInstances<NodeTaskTypeInfo>().Where(t => t.TaskTypeCode == taskType).Single();
            var bestTarget = GetBestTarget(taskTypeInfo);

            //TODO: Threshold check - decide if action is actually worth taking
            
            ApplyTask(taskTypeInfo, bestTarget);
        }

        private AINode GetBestTarget(NodeTaskTypeInfo task)
        {
            //Find the highest value node that can have the task assigned to it.
            AINode bestTarget = null;

            foreach (var aiNode in AINodes)
            {
                var gameNode = aiNode.GameNode;
                var checkState = task.CheckNodeValidity(gameNode, Player);

                if (checkState != TaskValidityState.Ok)
                    continue;

                if (bestTarget == null || aiNode.Value > bestTarget.Value)
                    bestTarget = aiNode;
            }

            return bestTarget;
        }

        private void ApplyTask(NodeTaskTypeInfo task, AINode target)
        {
            if (target == null)
                return;

            var worker = Player.UnassignedWorkers.FirstOrDefault();

            TempWorkerList.Clear();

            if (worker != null)
                TempWorkerList.Add(worker);

            var result = task.CheckTaskValidity(
                target.GameNode,
                Player,
                TempWorkerList,
                false);

            if (result != TaskValidityState.Ok)
                return;

            result = _playerActions.AddTask(
                    target.GameNode,
                    task,
                    TempWorkerList);

            if (result != TaskValidityState.Ok)
                return;
        }
    }
}
