﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Core.PlayerActions;
using Buildcraft.Core.NodeTasks.Effects;
using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.Core.PlayerTasks;
using Buildcraft.Core.PlayerTasks.Effects;
using Buildcraft.Core.Interfaces;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Diagnostics;
using Buildcraft.ScenarioBuilder;
using Buildcraft.Core.Interfaces.UnitTasks;
using Buildcraft.Core.UnitTasks;


namespace Buildcraft.Core.Test
{
    public abstract class BaseScenarioTest : IHostNotify
    {
        protected World world;

        protected WorldManager worldManager;
        protected PlayerManager playerManager;

        /// <summary>
        /// Container for dependency injection
        /// </summary>
        protected IBuildcraftContainer Container = new TestContainer();

        //Players
        protected Player player1;
        protected Player player2;

        //Player Capitals
        protected Node p1Capital;
        protected Node p2Capital;

        //Resource Types
        protected ResourceTypeInfo foodResource;
        protected ResourceTypeInfo moneyResource;

        //Improvement Types
        protected ImprovementTypeInfo cityImprovement;
        protected ImprovementTypeInfo farmImprovement;

        //Worker Types
        protected WorkerTypeInfo generalWorker;
        protected WorkerTypeInfo farmWorker;
        protected WorkerTypeInfo soldierWorker;

        //Worker Types
        protected UnitTypeInfo generalUnit;
        protected UnitTypeInfo farmUnit;
        protected UnitTypeInfo soldierUnit;

        //Task Types
        protected NodeTaskTypeInfo trainFarmerTask;
        protected NodeTaskTypeInfo trainSoldierTask;

        protected NodeTaskTypeInfo workImprovementTask;
        protected NodeTaskTypeInfo repairImprovementTask;
        protected NodeTaskTypeInfo patrolBordersTask;
        protected NodeTaskTypeInfo attackNodeTask;
        protected NodeTaskTypeInfo sabotageImprovementTask;

        //Unit Tasks
        protected IUnitTaskInfo workImprovementUnitTask;
        protected IUnitTaskInfo attackNodeUnitTask;
        protected IUnitTaskInfo defendRegionUnitTask;
        protected IUnitTaskInfo defendNodeUnitTask;

        //Player Tasks
        protected PlayerTaskTypeInfo workerConsumptionTask;
        protected PlayerTaskTypeInfo workerSpawnTask;

        //World Action
        protected TickWorldAction tickWorldAction;


        [TestInitialize]
        public void Initialize()
        {
            //Populate Container
            ContainerUtilities.LoadContainer(Container);
            Container.RegisterInstance<IHostNotify>(this);
            Container.RegisterInstance<ITaskScheduler>(new BasicTaskScheduler(100));

            //Initialize container
            WorldInitializer builder = new WorldInitializer();
            builder.InitializeContainer(Container);
            
            //Build Scenario
            ScenarioAssembler assembler = new ScenarioAssembler();
            assembler.BuildScenario(Container, ScenarioCodes.Line);

            //Unpack
            UnpackContainer();
            UnpackScenario();

            //Perform Bootstrapping
            Bootstrap();

            InitializationComplete();
        }



        private void UnpackScenario()
        {
            //Unpack Scenario
            player1 = world.Players[0];
            player2 = world.Players[1];

            p1Capital = player1.Nodes[0];
            p2Capital = player2.Nodes[0];
        }

        private void UnpackContainer()
        {
            //Load from Container
            world = Container.RetrieveInstance<World>();
            worldManager = Container.RetrieveInstance<WorldManager>();
            playerManager = Container.RetrieveInstance<PlayerManager>();

            //Resource Types            
            foodResource = Container.RetrieveInstance<ResourceTypeInfo>(ResourceTypeCodes.Food);
            moneyResource = Container.RetrieveInstance<ResourceTypeInfo>(ResourceTypeCodes.Money);

            //Improvement Types
            cityImprovement = Container.RetrieveInstance<ImprovementTypeInfo>(ImprovementTypeCodes.City);
            farmImprovement = Container.RetrieveInstance<ImprovementTypeInfo>(ImprovementTypeCodes.Farm);

            //Worker Types
            generalWorker = Container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Generalist);
            farmWorker = Container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Farmer);
            soldierWorker = Container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Soldier);

            //Task Types
            trainFarmerTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.TrainFarmer);
            trainSoldierTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.TrainSoldier);
            workImprovementTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.WorkImprovement);
            repairImprovementTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.RepairImprovement);
            patrolBordersTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.DefendTerritory);
            attackNodeTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.AttackNode);
            sabotageImprovementTask = Container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.SabotageImprovement);

            //Unit Types
            generalUnit = Container.RetrieveInstance<UnitTypeInfo>(WorkerTypeCodes.Generalist);
            farmUnit = Container.RetrieveInstance<UnitTypeInfo>(WorkerTypeCodes.Farmer);
            soldierUnit = Container.RetrieveInstance<UnitTypeInfo>(WorkerTypeCodes.Soldier);

            //Unit Task Types
            workImprovementUnitTask = Container.RetrieveInstance<IUnitTaskInfo>(UnitTaskTypeCodes.WorkImprovement);
            attackNodeUnitTask = Container.RetrieveInstance<IUnitTaskInfo>(UnitTaskTypeCodes.AttackNode);
            defendRegionUnitTask = Container.RetrieveInstance<IUnitTaskInfo>(UnitTaskTypeCodes.DefendRegion);
            defendNodeUnitTask = Container.RetrieveInstance<IUnitTaskInfo>(UnitTaskTypeCodes.DefendNode);

            //Player Types
            workerConsumptionTask = Container.RetrieveInstance<PlayerTaskTypeInfo>(PlayerTaskTypeCodes.WorkerConsumption);
            workerSpawnTask = Container.RetrieveInstance<PlayerTaskTypeInfo>(PlayerTaskTypeCodes.WorkerSpawn);

            //World Actions
            tickWorldAction = Container.RetrieveInstance<TickWorldAction>();
        }

        private void Bootstrap()
        {
            //TODO: Boot strap the task types
        }


        protected abstract void InitializationComplete();

        protected Player CreatePlayer(string playerName)
        {
            //Create Player
            var player = worldManager.AddPlayer(playerName);

            //Add Default tasks, will possibly need to be extended.
            playerManager.AddPlayerTask(player, workerSpawnTask);
            playerManager.AddPlayerTask(player, workerConsumptionTask);

            return player;
        }


        #region IHostNotify Members

        public virtual void TaskAdded(NodeTask task)
        {
            Trace.WriteLine(string.Format("Task Added: {0}", task.ToString()));
        }

        public virtual void TaskRemoved(NodeTask task)
        {
            Trace.WriteLine(string.Format("Task Removed: {0}", task.ToString()));
        }

        public void ResourcesProduced(NodeTask task, ResourceTypeInfo resourceType, int production)
        {
            Trace.WriteLine(string.Format("{0} production - {1} : {2}", task.Owner.Name, resourceType, production));
        }

        public void GameOver(Player winner)
        {
            Trace.WriteLine(string.Format("GAME OVER! Player {0} wins.", winner.Name));
        }

        #endregion
    }

}
