﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Buildcraft.Core.Components;
using Buildcraft.Core.Model;

using Microsoft.Practices.Unity;
using Buildcraft.Core.Interfaces;
using System.Diagnostics;
using Buildcraft.ScenarioBuilder;

namespace Buildcraft.Core.Test
{
    public abstract class BaseTest : IHostNotify
    {
        protected World world;

        protected WorldManager worldManager;
        protected PlayerManager playerManager;

        /// <summary>
        /// Container for dependency injection
        /// </summary>
        protected IBuildcraftContainer Container = new TestContainer();

        [TestInitialize]
        public void Initialize()
        {
            //Populate Container
            ContainerUtilities.LoadContainer(Container);
            Container.RegisterInstance<IHostNotify>(HostNotifyMock);
            Container.RegisterInstance<ITaskScheduler>(TaskSchedulerMock);

            //Load from Container
            world = Container.RetrieveInstance<World>();
            worldManager = Container.RetrieveInstance<WorldManager>();
            playerManager = Container.RetrieveInstance<PlayerManager>();

            CustomInitialize();
            Container.Initialize();
            ConfigureWorld();
        }

        private IHostNotify _hostNotifyMock = Rhino.Mocks.MockRepository.GenerateStub<IHostNotify>();
        protected virtual IHostNotify HostNotifyMock
        {
            get { return _hostNotifyMock; }
        }

        private ITaskScheduler _taskSchedulerMock = Rhino.Mocks.MockRepository.GenerateStub<ITaskScheduler>();
        protected virtual ITaskScheduler TaskSchedulerMock
        {
            get { return _taskSchedulerMock; }
        }


        protected abstract void CustomInitialize();

        protected abstract void ConfigureWorld();


        protected virtual Player CreatePlayer(string playerName)
        {
            //Create Player
            var player = worldManager.AddPlayer(playerName);

            //Anything else?

            return player;
        }

        protected virtual Node CreateNode(Player owner = null, ImprovementTypeInfo typeInfo = null)
        {
            //Create Node, assign ownership and set improvement
            var node = worldManager.AddNode();

            if (typeInfo != null)
                worldManager.SetNodeImprovement(node, typeInfo);

            if (owner != null)
                worldManager.SetNodeOwner(node, owner);

            return node;
        }

        protected virtual void CreateJoin(Node from, Node to)
        {
            worldManager.AddJoin(from, to);
        }


        #region IHostNotify Members

        public virtual void TaskAdded(NodeTask task)
        {
            Trace.WriteLine(string.Format("Task Added: {0}", task.ToString()));
        }

        public virtual void TaskRemoved(NodeTask task)
        {
            Trace.WriteLine(string.Format("Task Removed: {0}", task.ToString()));
        }

        public void ResourcesProduced(NodeTask task, ResourceTypeInfo resourceType, int production)
        {
            Trace.WriteLine(string.Format("{0} production - {1} : {2}", task.Owner.Name, resourceType, production));
        }

        public void GameOver(Player winner)
        {
            Trace.WriteLine(string.Format("GAME OVER! Player {0} wins.", winner.Name));
        }

        #endregion
    }
}
