﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Buildcraft.Core.Components;
using Buildcraft.Core.Model;

using Buildcraft.ScenarioBuilder;
using Buildcraft.Core.Interfaces.UnitTasks.TaskInfo;
using Buildcraft.Core.Interfaces.UnitTasks;
using Rhino.Mocks;
using Buildcraft.Core.UnitTasks;

namespace Buildcraft.Core.Test.ComponentTests
{
    [TestClass]
    public class PlayerManagerTests : BaseTest
    {
        private Player _player;

        protected override void CustomInitialize()
        {
        }

        protected override void ConfigureWorld()
        {
            _player = worldManager.AddPlayer("P1");
        }


        [TestMethod]
        public void PlayerManagerTests_AddRemoveWorker()
        {
            var player1 = worldManager.AddPlayer("P1");
            var typeInfo = new WorkerTypeInfo(WorkerTypeCodes.Generalist);

            var worker1 = playerManager.AddWorker(player1, typeInfo);

            Assert.IsTrue(worker1.Owner == player1);
            Assert.IsTrue(worker1.TypeInfo == typeInfo);
            Assert.IsTrue(worker1.Id > 0);

            Assert.IsTrue(player1.AllWorkers.Count == 1);
            Assert.IsTrue(player1.AllWorkers[0] == worker1);
            Assert.IsTrue(player1.UnassignedWorkers.Count == 1);
            Assert.IsTrue(player1.UnassignedWorkers[0] == worker1);


            var worker2 = playerManager.AddWorker(player1, typeInfo);

            Assert.IsTrue(player1.AllWorkers.Count ==2 );
            Assert.IsTrue(player1.AllWorkers[1] == worker2);
            Assert.IsTrue(player1.UnassignedWorkers.Count == 2);
            Assert.IsTrue(player1.UnassignedWorkers[1] == worker2);

            playerManager.RemoveWorker(worker1);

            Assert.IsTrue(player1.AllWorkers.Count == 1);
            Assert.IsTrue(player1.AllWorkers[0] == worker2);
            Assert.IsTrue(player1.UnassignedWorkers.Count == 1);
            Assert.IsTrue(player1.UnassignedWorkers[0] == worker2);
        }

        [TestMethod]
        public void PlayerManagerTests_AddRemoveUnit()
        {
            var player1 = worldManager.AddPlayer("P1");
            var typeInfo = new UnitTypeInfo(WorkerTypeCodes.Generalist);

            var unit1 = playerManager.AddUnit(player1, typeInfo);

            Assert.IsTrue(unit1.Owner == player1);
            Assert.IsTrue(unit1.TypeInfo == typeInfo);
            Assert.IsTrue(unit1.Id > 0);

            Assert.IsTrue(player1.AllUnits.Count == 1);
            Assert.IsTrue(player1.AllUnits[0] == unit1);
            Assert.IsTrue(player1.UnassignedUnits.Count == 1);
            Assert.IsTrue(player1.UnassignedUnits[0] == unit1);


            var unit2 = playerManager.AddUnit(player1, typeInfo);

            Assert.IsTrue(player1.AllUnits.Count == 2);
            Assert.IsTrue(player1.AllUnits[1] == unit2);
            Assert.IsTrue(player1.UnassignedUnits.Count == 2);
            Assert.IsTrue(player1.UnassignedUnits[1] == unit2);
            

            playerManager.RemoveUnit(unit1);

            Assert.IsTrue(player1.AllUnits.Count == 1);
            Assert.IsTrue(player1.AllUnits[0] == unit2);
            Assert.IsTrue(player1.UnassignedUnits.Count == 1);
            Assert.IsTrue(player1.UnassignedUnits[0] == unit2);
        }


        [TestMethod]
        public void PlayerManagerTests_AddRemoveTask()
        {
            var player1 = worldManager.AddPlayer("P1");
            var node = worldManager.AddNode();
            var taskInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.WorkImprovement, null, null);


            var task1 = playerManager.AddNodeTask(player1, node, taskInfo);

            Assert.IsTrue(task1.TaskId > 0);
            Assert.IsTrue(task1.Node == node);
            Assert.IsTrue(task1.Owner == player1);

            Assert.IsTrue(node.Tasks.Count == 1);
            Assert.IsTrue(node.Tasks[0] == task1);

            Assert.IsTrue(player1.NodeTasks.Count == 1);
            Assert.IsTrue(player1.NodeTasks[0] == task1);


            var task2 = playerManager.AddNodeTask(player1, node, taskInfo);

            Assert.IsTrue(task2.TaskId > 1);
            Assert.IsTrue(task2.Node == node);
            Assert.IsTrue(task2.Owner == player1);

            Assert.IsTrue(node.Tasks.Count == 2);
            Assert.IsTrue(node.Tasks[1] == task2);

            Assert.IsTrue(player1.NodeTasks.Count == 2);
            Assert.IsTrue(player1.NodeTasks[1] == task2);


            playerManager.RemoveNodeTask(task1);

            Assert.IsTrue(node.Tasks.Count == 1);
            Assert.IsTrue(node.Tasks[0] == task2);

            Assert.IsTrue(player1.NodeTasks.Count == 1);
            Assert.IsTrue(player1.NodeTasks[0] == task2);
        }

        [TestMethod]
        public void PlayerManagerTests_AssignUnassignClearWorkers()
        {
            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.WorkImprovement, null, null);
            var workerTypeInfo = new WorkerTypeInfo(WorkerTypeCodes.Generalist);

            var node = worldManager.AddNode();

            var task1 = playerManager.AddNodeTask(_player, node, taskTypeInfo);
            var worker1 = playerManager.AddWorker(_player, workerTypeInfo);
            var worker2 = playerManager.AddWorker(_player, workerTypeInfo);


            playerManager.AssignWorker(task1, worker1);

            Assert.IsTrue(_player.UnassignedWorkers.Count == 1);

            Assert.IsTrue(task1.Workers.Count == 1);
            Assert.IsTrue(task1.Workers[0] == worker1);


            playerManager.AssignWorker(task1, worker2);

            Assert.IsTrue(_player.UnassignedWorkers.Count == 0);

            Assert.IsTrue(task1.Workers.Count == 2);
            Assert.IsTrue(task1.Workers[1] == worker2);


            playerManager.UnassignWorker(task1, worker1);

            Assert.IsTrue(_player.UnassignedWorkers.Count == 1);

            Assert.IsTrue(task1.Workers.Count == 1);
            Assert.IsTrue(task1.Workers[0] == worker2);


            playerManager.ClearWorkers(task1);

            Assert.IsTrue(_player.UnassignedWorkers.Count == 2);
            Assert.IsTrue(task1.Workers.Count == 0);
        }


        [TestMethod]
        public void PlayerManagerTest_AssignUnassignOneUnitOneTask()
        {
            //Create Task Parts
            IUnitTaskInfo taskInfo = MockRepository.GenerateMock<IUnitTaskInfo>();
            IUnitTaskStateItem taskState = MockRepository.GenerateMock<IUnitTaskStateItem>();
            IUnitTaskStateBucket taskStateBucket = MockRepository.GenerateMock<IUnitTaskStateBucket>();
            List<IUnitTaskStateItem> bucketState = new List<IUnitTaskStateItem>();

            //Create unit and node
            var node = worldManager.AddNode();
            var unitTypeInfo = new UnitTypeInfo("Unit");
            var unit = playerManager.AddUnit(_player, unitTypeInfo);

            //Stub TaskInfo
            taskInfo.Stub(t => t.CreateTaskState(null, null)).IgnoreArguments().Return(taskState);
            taskInfo.Stub(t => t.CreateTaskStateBucket(null, null)).IgnoreArguments().Return(taskStateBucket);

            //Stub Task State
            taskState.Stub(t => t.TaskInfo).Return(taskInfo);

            //Stub TaskStateBucket
            taskStateBucket.Stub(t => t.Node).Return(node);
            taskStateBucket.Stub(t => t.Owner).Return(_player);
            taskStateBucket.Stub(t => t.TaskInfo).Return(taskInfo);
            taskStateBucket.Stub(t => t.TaskStates).Return(bucketState);


            //STEP1: Assign unit to task
            playerManager.AssignUnit(node, unit, taskInfo);

            //Check Get State was called corectly
            taskInfo.AssertWasCalled(t => t.CreateTaskState(
                    Arg<Node>.Is.Equal(node),
                    Arg<Unit>.Is.Equal(unit)));

            //Check Get State Bucket was called correctly
            taskInfo.AssertWasCalled(t => t.CreateTaskStateBucket(
                    Arg<Node>.Is.Equal(node),
                    Arg<Player>.Is.Equal(_player)));

            //Check bucket and unit were registered with scheduler
            base.TaskSchedulerMock.AssertWasCalled(t => t.RegisterStateBucket(Arg<IUnitTaskStateBucket>.Is.Equal(taskStateBucket)));
            base.TaskSchedulerMock.AssertWasCalled(t => t.RegisterState(Arg<IUnitTaskStateItem>.Is.Equal(taskState)));

            //Check Bucket exists on node
            Assert.IsTrue(node.TaskStateBuckets.Count == 1);
            Assert.IsTrue(node.TaskStateBuckets[0] == taskStateBucket);

            //Check Unit on node
            Assert.IsTrue(node.Units.Count == 1);
            Assert.IsTrue(node.Units[0] == unit);

            //Check Unit correctly assigned Tasks
            Assert.IsTrue(unit.TaskState == taskState);


            //STEP2: Unassign unit from task
            playerManager.UnassignUnit(unit);

            //Check Node / Units Cleared
            Assert.IsTrue(node.Units.Count == 0);
            Assert.IsTrue(node.TaskStateBuckets.Count == 0);
            Assert.IsTrue(unit.TaskState == null);

            //Check bucket and unit were unregistered with scheduler
            base.TaskSchedulerMock.AssertWasCalled(t => t.UnregisterStateBucket(Arg<IUnitTaskStateBucket>.Is.Equal(taskStateBucket)));
            base.TaskSchedulerMock.AssertWasCalled(t => t.UnregisterState(Arg<IUnitTaskStateItem>.Is.Equal(taskState)));

        }

        [TestMethod]
        public void PlayerManagerTest_AssignUnassignTwoUnitsOneTask()
        {
            //Create Task Parts
            IUnitTaskInfo taskInfo = MockRepository.GenerateMock<IUnitTaskInfo>();
            IUnitTaskStateItem taskState = MockRepository.GenerateMock<IUnitTaskStateItem>();
            IUnitTaskStateBucket taskStateBucket = MockRepository.GenerateMock<IUnitTaskStateBucket>();
            List<IUnitTaskStateItem> bucketState = new List<IUnitTaskStateItem>();

            //Create units and node
            var node = worldManager.AddNode();
            var unitTypeInfo = new UnitTypeInfo("Unit");

            var unit1 = playerManager.AddUnit(_player, unitTypeInfo);
            var unit2 = playerManager.AddUnit(_player, unitTypeInfo);

            //Stub TaskInfo
            taskInfo.Stub(t => t.CreateTaskState(null, null)).IgnoreArguments().Return(taskState);
            taskInfo.Stub(t => t.CreateTaskStateBucket(null, null)).IgnoreArguments().Return(taskStateBucket);

            //Stub Task State
            taskState.Stub(t => t.TaskInfo).Return(taskInfo);

            //Stub TaskStateBucket
            taskStateBucket.Stub(t => t.Node).Return(node);
            taskStateBucket.Stub(t => t.Owner).Return(_player);
            taskStateBucket.Stub(t => t.TaskInfo).Return(taskInfo);
            taskStateBucket.Stub(t => t.TaskStates).Return(bucketState);
            taskStateBucket.Stub(t => t.AddTaskState(null)).IgnoreArguments().Callback((IUnitTaskStateItem state, bool recalc) =>
            {
                taskStateBucket.TaskStates.Add(state);
                return true;
            });
            taskStateBucket.Stub(t => t.RemoveTaskState(null)).IgnoreArguments().Callback((IUnitTaskStateItem state, bool recalc) =>
            {
                taskStateBucket.TaskStates.Remove(state);
                return true;
            });

            //STEP 1.1: Assign first unit to task
            playerManager.AssignUnit(node, unit1, taskInfo);

            //Check Get State was called corectly
            taskInfo.AssertWasCalled(t => t.CreateTaskState(
                    Arg<Node>.Is.Equal(node),
                    Arg<Unit>.Is.Equal(unit1)));

            //Check Get State Bucket was called correctly
            taskInfo.AssertWasCalled(t => t.CreateTaskStateBucket(
                    Arg<Node>.Is.Equal(node),
                    Arg<Player>.Is.Equal(_player)));

            //Check Bucket exists on node
            Assert.IsTrue(node.TaskStateBuckets.Count == 1);
            Assert.IsTrue(node.TaskStateBuckets[0] == taskStateBucket);

            //Check Unit on node
            Assert.IsTrue(node.Units.Count == 1);
            Assert.IsTrue(node.Units[0] == unit1);

            //Check Unit correctly assigned Tasks
            Assert.IsTrue(unit1.TaskState == taskState);


            //STEP 1.2: Assign second unit to task
            playerManager.AssignUnit(node, unit2, taskInfo);

            //Check Get State was called corectly
            taskInfo.AssertWasCalled(t => t.CreateTaskState(
                    Arg<Node>.Is.Equal(node),
                    Arg<Unit>.Is.Equal(unit2)));

            //Check Bucket exists on node
            Assert.IsTrue(node.TaskStateBuckets.Count == 1);
            Assert.IsTrue(node.TaskStateBuckets[0] == taskStateBucket);

            //Check Unit on node
            Assert.IsTrue(node.Units.Count == 2);
            Assert.IsTrue(node.Units[1] == unit2);

            //Check Unit correctly assigned Tasks
            Assert.IsTrue(unit2.TaskState == taskState);


            //STEP 2.1: Unassign first unit from task
            playerManager.UnassignUnit(unit1);

            //Check Unit Removed
            Assert.IsTrue(node.Units.Count == 1);
            Assert.IsTrue(node.TaskStateBuckets.Count == 1);
            Assert.IsTrue(unit1.TaskState == null);


            //STEP 2.2: Unassign second unit from task
            playerManager.UnassignUnit(unit2);

            //Check Unit Removed
            Assert.IsTrue(node.Units.Count == 0);
            Assert.IsTrue(node.TaskStateBuckets.Count == 0);
            Assert.IsTrue(unit2.TaskState == null);
        }

        [TestMethod]
        public void PlayerManagerTest_AssignUnassignTwoUnitsTwoTasks()
        {
            //Create Task Part 1
            IUnitTaskInfo taskInfo1 = MockRepository.GenerateMock<IUnitTaskInfo>();
            IUnitTaskStateItem taskState1 = MockRepository.GenerateMock<IUnitTaskStateItem>();
            IUnitTaskStateBucket taskStateBucket1 = MockRepository.GenerateMock<IUnitTaskStateBucket>();
            List<IUnitTaskStateItem> taskStateBucket1State = new List<IUnitTaskStateItem>();

            //Stub TaskInfo
            taskInfo1.Stub(t => t.CreateTaskState(null, null)).IgnoreArguments().Return(taskState1);
            taskInfo1.Stub(t => t.CreateTaskStateBucket(null, null)).IgnoreArguments().Return(taskStateBucket1);

            //Stub Task State
            taskState1.Stub(t => t.TaskInfo).Return(taskInfo1);


            //Create Task Part 2
            IUnitTaskInfo taskInfo2 = MockRepository.GenerateMock<IUnitTaskInfo>();
            IUnitTaskStateItem taskState2 = MockRepository.GenerateMock<IUnitTaskStateItem>();
            IUnitTaskStateBucket taskStateBucket2 = MockRepository.GenerateMock<IUnitTaskStateBucket>();
            List<IUnitTaskStateItem> taskStateBucket2State = new List<IUnitTaskStateItem>();

            //Stub TaskInfo 2
            taskInfo2.Stub(t => t.CreateTaskState(null, null)).IgnoreArguments().Return(taskState2);
            taskInfo2.Stub(t => t.CreateTaskStateBucket(null, null)).IgnoreArguments().Return(taskStateBucket2);

            //Stub Task State 2
            taskState2.Stub(t => t.TaskInfo).Return(taskInfo2);


            //Create unit and node
            var node = worldManager.AddNode();
            var unitTypeInfo = new UnitTypeInfo("Unit");

            var unit1 = playerManager.AddUnit(_player, unitTypeInfo);
            var unit2 = playerManager.AddUnit(_player, unitTypeInfo);

            //Stub TaskStateBucket 1
            taskStateBucket1.Stub(t => t.Node).Return(node);
            taskStateBucket1.Stub(t => t.Owner).Return(_player);
            taskStateBucket1.Stub(t => t.TaskInfo).Return(taskInfo1);
            taskStateBucket1.Stub(t => t.TaskStates).Return(taskStateBucket1State);
            taskStateBucket1.Stub(t => t.AddTaskState(null)).IgnoreArguments().Callback((IUnitTaskStateItem state, bool recalc) =>
            {
                taskStateBucket1.TaskStates.Add(state);
                return true;
            });
            taskStateBucket1.Stub(t => t.RemoveTaskState(null)).IgnoreArguments().Callback((IUnitTaskStateItem state, bool recalc) =>
            {
                taskStateBucket1.TaskStates.Remove(state);
                return true;
            });

            //Stub TaskStateBucket 2
            taskStateBucket2.Stub(t => t.Node).Return(node);
            taskStateBucket2.Stub(t => t.Owner).Return(_player);
            taskStateBucket2.Stub(t => t.TaskInfo).Return(taskInfo2);
            taskStateBucket2.Stub(t => t.TaskStates).Return(taskStateBucket2State);
            taskStateBucket2.Stub(t => t.AddTaskState(null)).IgnoreArguments().Callback((IUnitTaskStateItem state, bool recalc) =>
            {
                taskStateBucket2.TaskStates.Add(state);
                return true;
            });
            taskStateBucket2.Stub(t => t.RemoveTaskState(null)).IgnoreArguments().Callback((IUnitTaskStateItem state, bool recalc) =>
            {
                taskStateBucket2.TaskStates.Remove(state);
                return true;
            });


            //STEP 1.1: Assign first unit to task
            playerManager.AssignUnit(node, unit1, taskInfo1);

            //Check Get State was called corectly
            taskInfo1.AssertWasCalled(t => t.CreateTaskState(
                    Arg<Node>.Is.Equal(node),
                    Arg<Unit>.Is.Equal(unit1)));

            //Check Get State Bucket was called correctly
            taskInfo1.AssertWasCalled(t => t.CreateTaskStateBucket(
                    Arg<Node>.Is.Equal(node),
                    Arg<Player>.Is.Equal(_player)));

            //Check Bucket exists on node
            Assert.IsTrue(node.TaskStateBuckets.Count == 1);
            Assert.IsTrue(node.TaskStateBuckets[0] == taskStateBucket1);

            //Check Unit on node
            Assert.IsTrue(node.Units.Count == 1);
            Assert.IsTrue(node.Units[0] == unit1);

            //Check Unit correctly assigned Tasks
            Assert.IsTrue(unit1.TaskState == taskState1);


            //STEP 1.2: Assign second unit to task
            playerManager.AssignUnit(node, unit2, taskInfo2);

            //Check Get State was called corectly
            taskInfo2.AssertWasCalled(t => t.CreateTaskState(
                    Arg<Node>.Is.Equal(node),
                    Arg<Unit>.Is.Equal(unit2)));

            //Check Bucket exists on node
            Assert.IsTrue(node.TaskStateBuckets.Count == 2);
            Assert.IsTrue(node.TaskStateBuckets[1] == taskStateBucket2);

            //Check Unit on node
            Assert.IsTrue(node.Units.Count == 2);
            Assert.IsTrue(node.Units[1] == unit2);

            //Check Unit correctly assigned Tasks
            Assert.IsTrue(unit2.TaskState == taskState2);


            //STEP 2.1: Unassign first unit from task
            playerManager.UnassignUnit(unit1);

            //Check Unit Removed
            Assert.IsTrue(node.Units.Count == 1);
            Assert.IsTrue(node.TaskStateBuckets.Count == 1);
            Assert.IsTrue(unit1.TaskState == null);


            //STEP 2.2: Unassign second unit from task
            playerManager.UnassignUnit(unit2);

            //Check Unit Removed
            Assert.IsTrue(node.Units.Count == 0);
            Assert.IsTrue(node.TaskStateBuckets.Count == 0);
            Assert.IsTrue(unit2.TaskState == null);
        }
    }
}
