﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.Core.Model;

using Buildcraft.ScenarioBuilder;

namespace Buildcraft.Core.Test.ConditionTests
{
    [TestClass]
    public class TrainWorkerConditionTests : BaseTest
    {
        //Model Parts
        private Node node;
        private Player player;

        //TypeInfo's
        private ImprovementTypeInfo improvementTypeInfo;
        private ImprovementTypeInfo invalidImprovementTypeInfo;

        private WorkerTypeInfo generalWorkerTypeInfo;
        private WorkerTypeInfo farmerWorkerTypeInfo;

        //Condition
        TrainWorkerCondition condition = new TrainWorkerCondition();


        protected override void CustomInitialize()
        {
        }

        protected override void ConfigureWorld()
        {
            improvementTypeInfo = new ImprovementTypeInfo(ImprovementTypeCodes.City);
            invalidImprovementTypeInfo = new ImprovementTypeInfo(ImprovementTypeCodes.Farm);
            generalWorkerTypeInfo = new WorkerTypeInfo(WorkerTypeCodes.Generalist);
            farmerWorkerTypeInfo = new WorkerTypeInfo(WorkerTypeCodes.Farmer);

            node = worldManager.AddNode();
            player = worldManager.AddPlayer("P1");

            condition.TargetWorkerType = farmerWorkerTypeInfo;
            condition.MinWorkers = 1;
            condition.MaxWorkers = 1;

            worldManager.SetNodeOwner(node, player);
        }

        [TestMethod]
        public void TrainWorkerConditionTests_Basic()
        {
            CheckNodeValidity();
            CheckWorkerValidity();
            CheckWorkersValidity();
        }

        private void CheckNodeValidity()
        {
            var player2 = worldManager.AddPlayer("P2");

            //Check No Improvement State
            var result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.NoImprovement, result);

            //Check Valid For Node
            worldManager.SetNodeImprovement(node, improvementTypeInfo);
            result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.Ok, result);

            //Check Invalid For Enemy
            worldManager.SetNodeOwner(node, player2);
            result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.NotAllowedInEnemyTerritory, result);
            worldManager.SetNodeOwner(node, player);

            //Check Invalid For Improvement
            condition.ImprovementRestriction = invalidImprovementTypeInfo;
            result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.WrongImprovement, result);
            condition.ImprovementRestriction = null;

        }

        private void CheckWorkerValidity()
        {
            var worker = playerManager.AddWorker(player, generalWorkerTypeInfo);

            //Check With Invalid Worker
            var result = condition.CheckWorkerValidity(node, player, worker);
            Assert.AreEqual(TaskValidityState.Ok, result);

            //Check Worker Type when different from parameter (Valid)
            worker.TypeInfo = farmerWorkerTypeInfo;
            result = condition.CheckWorkerValidity(node, player, worker);
            Assert.AreEqual(TaskValidityState.InvalidForWorker, result);
        }

        private void CheckWorkersValidity()
        {
            var workers = new List<Worker>();

            //Check too Few
            var result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.TooFewWorkers, result);

            //Check Invalid worker
            workers.Add(playerManager.AddWorker(player, generalWorkerTypeInfo));
            workers[0].TypeInfo = farmerWorkerTypeInfo;
            result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.OneOrMoreWorkerIsInvalid, result);
            workers[0].TypeInfo = generalWorkerTypeInfo;

            //Check Ok
            result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.Ok, result);

            //Check too Many
            workers.Add(playerManager.AddWorker(player, generalWorkerTypeInfo));
            result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.TooManyWorkers, result);
        }

    }
}
