﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Buildcraft.Core.Model;

using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.ScenarioBuilder;

namespace Buildcraft.Core.Test.ConditionTests
{
    [TestClass]
    public class WorkImprovementConditionTests : BaseTest
    {
        //Model Parts
        private Node node;
        private Player player;
        private Player player2;
        private Worker worker;

        //TypeInfo's
        private ImprovementTypeInfo improvementTypeInfo;
        private WorkerTypeInfo workerTypeInfo;
        private WorkerTypeInfo invalidWorkerTypeInfo;

        //Condition
        WorkImprovementCondition condition = new WorkImprovementCondition();

        protected override void CustomInitialize()
        {
        }

        protected override void ConfigureWorld()
        {
            improvementTypeInfo = new ImprovementTypeInfo(ImprovementTypeCodes.City);
            workerTypeInfo = new WorkerTypeInfo(WorkerTypeCodes.Generalist);
            invalidWorkerTypeInfo = new WorkerTypeInfo(WorkerTypeCodes.Administrator);

            improvementTypeInfo.WorkerTypeRestriction = workerTypeInfo;
            improvementTypeInfo.MinWorkers = 1;
            improvementTypeInfo.MaxWorkers = 1;

            node = worldManager.AddNode();
            player = worldManager.AddPlayer("P1");
            player2 = worldManager.AddPlayer("P2");
            worker = playerManager.AddWorker(player, workerTypeInfo);

            worldManager.SetNodeOwner(node, player);
        }

        [TestMethod]
        public void WorkImprovementConditionTests_Basic()
        {
            CheckNodeValidity();
            CheckWorkerValidity();
            CheckWorkersValidity();
        }

        private void CheckNodeValidity()
        {
            //Check No Improvement State
            var result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.NoImprovement, result);

            //Check Valid For Node
            worldManager.SetNodeImprovement(node, improvementTypeInfo);
            result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.Ok, result);

            //Check Invalid For Enemy
            worldManager.SetNodeOwner(node, player2);
            result = condition.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.NotAllowedInEnemyTerritory, result);

        }

        private void CheckWorkerValidity()
        {
            //Worker Validity Setup
            worldManager.SetNodeOwner(node, player);
            improvementTypeInfo.WorkerTypeRestriction = invalidWorkerTypeInfo;

            //Check Worker Validity
            var result = condition.CheckWorkerValidity(node, player, worker);
            Assert.AreEqual(TaskValidityState.WorkerIsWrongType, result);

            //Check Worker Type is Specific
            improvementTypeInfo.WorkerTypeRestriction = workerTypeInfo;
            result = condition.CheckWorkerValidity(node, player, worker);
            Assert.AreEqual(TaskValidityState.Ok, result);

            //Check Worker Type when generic
            improvementTypeInfo.WorkerTypeRestriction = null;
            result = condition.CheckWorkerValidity(node, player, worker);
            Assert.AreEqual(TaskValidityState.Ok, result);
        }

        private void CheckWorkersValidity()
        {
            var workers = new List<Worker>();

            //Check Too Few
            var result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.TooFewWorkers, result);

            //Check With Valid Worker
            workers.Add(worker);
            result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.Ok, result);

            //Check With Invalid Worker
            improvementTypeInfo.WorkerTypeRestriction = workerTypeInfo;
            worker.TypeInfo = invalidWorkerTypeInfo;
            result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.OneOrMoreWorkerIsInvalid, result);
            worker.TypeInfo = workerTypeInfo;

            //Check Too Many
            workers.Add(playerManager.AddWorker(player, workerTypeInfo));
            result = condition.CheckWorkersValidity(node, player, workers);
            Assert.AreEqual(TaskValidityState.TooManyWorkers, result);
        }
    }
}
