﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Buildcraft.Core.Model;
using Buildcraft.Core.Components;

using Buildcraft.Core.NodeTasks.Effects;
using Buildcraft.ScenarioBuilder;

namespace Buildcraft.Core.Test.NodeEffectTests
{
    [TestClass]
    public class NodeNodeEffectTests : BaseTest
    {
        private Player player;
        private Node node;

        private ResourceTypeInfo foodResourceType;
        private ResourceTypeInfo moneyResourceType;

        protected override void CustomInitialize()
        {
        }

        protected override void ConfigureWorld()
        {
            foodResourceType = new ResourceTypeInfo(ResourceTypeCodes.Food);
            moneyResourceType = new ResourceTypeInfo(ResourceTypeCodes.Money);

            player = worldManager.AddPlayer("P1");
            node = worldManager.AddNode();
            worldManager.SetNodeOwner(node, player);
        }


        [TestMethod]
        public void NodeEffectTests_ResourceProduction()
        {
            var farmerType = new WorkerTypeInfo(WorkerTypeCodes.Farmer);
            var generalType = new WorkerTypeInfo(WorkerTypeCodes.Generalist);

            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.AttackNode, null, null);


            ResourceProductionEffect effect = new ResourceProductionEffect();
            effect.Initialize(Container);
            effect.ResourceProduced = foodResourceType;
            effect.ProductionPerWorker = 1;
            effect.SpecialistMultiplier = 2;
            effect.SpecialistType = farmerType;

            var farmer = playerManager.AddWorker(player, farmerType);
            var worker = playerManager.AddWorker(player, generalType);

            NodeTask task = playerManager.AddNodeTask(player, node, taskTypeInfo);
            task.Owner = player;

            playerManager.AssignWorker(task, worker);
            playerManager.AssignWorker(task, farmer);

            effect.Apply(task); //No Effect
            effect.Fire(task); //Fire the task
            effect.Remove(task); //No Effect

            Assert.AreEqual(3, player.Resources[foodResourceType]);
        }

        [TestMethod]
        public void NodeEffectTests_AttackNode()
        {
            var soldierType = new WorkerTypeInfo(WorkerTypeCodes.Soldier);
            var generalType = new WorkerTypeInfo(WorkerTypeCodes.Generalist);

            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.AttackNode, null, null);


            Player player2 = worldManager.AddPlayer("P2");
            worldManager.SetNodeOwner(node, player2);
            player2.Defence = 5;

            AttackNodeEffect effect = new AttackNodeEffect();
            effect.AttackPerWorker = 1;
            effect.SpecialistType = soldierType;
            effect.SpecialistMultiplier = 5;
            effect.Initialize(Container);

            var soldier = playerManager.AddWorker(player, soldierType);
            var worker = playerManager.AddWorker(player, generalType);

            NodeTask task = playerManager.AddNodeTask(player, node, taskTypeInfo);
            playerManager.AssignWorker(task, worker);

            Assert.AreEqual(player2, node.Owner);

            effect.Apply(task); //No Effect
            effect.Fire(task); //Fire the task
            effect.Remove(task); //No Effect

            playerManager.AssignWorker(task, soldier);
            Assert.AreEqual(player2, node.Owner);

            effect.Apply(task); //No Effect
            effect.Fire(task); //Fire the task
            effect.Remove(task); //No Effect

            Assert.AreEqual(player, node.Owner);
        }

        [TestMethod]
        public void NodeEffectTests_DefendTerritory()
        {
            //Check that the effect correctly calculates the defence value, provided by the city.
            var soldierType = new WorkerTypeInfo(WorkerTypeCodes.Soldier);
            var generalType = new WorkerTypeInfo(WorkerTypeCodes.Generalist);

            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.AttackNode, null, null);

            DefendTerritoryEffect effect = new DefendTerritoryEffect();
            effect.DefencePerWorker = 1;
            effect.SpecialistType = soldierType;
            effect.SpecialistMultiplier = 5;
            effect.Initialize(Container);

            var soldier = playerManager.AddWorker(player, soldierType);
            var worker = playerManager.AddWorker(player, generalType);

            NodeTask task = playerManager.AddNodeTask(player, node, taskTypeInfo);

            playerManager.AssignWorker(task, worker);
            playerManager.AssignWorker(task, soldier);

            //Calculate defence check
            effect.Apply(task); 
            Assert.AreEqual(6, player.Defence);

            //Remove unit and refresh Check
            playerManager.UnassignWorker(task, worker);
            effect.Refresh(task); 
            Assert.AreEqual(5, player.Defence);

            //Test no effect (on defence)
            effect.Fire(task); 
            Assert.AreEqual(5, player.Defence);

            //Decrease Realm Defence
            effect.Remove(task); 
            Assert.AreEqual(0, player.Defence);
        }

        [TestMethod]
        public void NodeEffectTests_ChangeWorkerType()
        {
            //Check that the effect correctly calculates the defence value, provided by the city.
            var soldierType = new WorkerTypeInfo(WorkerTypeCodes.Soldier);
            var generalType = new WorkerTypeInfo(WorkerTypeCodes.Generalist);

            Container.RegisterInstance(soldierType, WorkerTypeCodes.Soldier);
            Container.RegisterInstance(generalType, WorkerTypeCodes.Generalist);

            ChangeWorkerTypeEffect effect = new ChangeWorkerTypeEffect(WorkerTypeCodes.Soldier);
            effect.Initialize(Container);

            var worker1 = playerManager.AddWorker(player, generalType);
            var worker2 = playerManager.AddWorker(player, generalType);

            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.TrainFarmer, null, null);
            NodeTask task = playerManager.AddNodeTask(player, node, taskTypeInfo);


            playerManager.AssignWorker(task, worker1);
            playerManager.AssignWorker(task, worker2);

            effect.Apply(task); //No Effect

            effect.Fire(task); //Fire the task
            Assert.AreEqual(soldierType, worker1.TypeInfo);
            Assert.AreEqual(soldierType, worker2.TypeInfo);

            effect.Remove(task); //No Effect
        }

        [TestMethod]
        public void NodeEffectTests_SabotageAndRepairImprovement()
        {
            var improvementType = new ImprovementTypeInfo(ImprovementTypeCodes.Farm);
            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.WorkImprovement, null, null);
            var task = playerManager.AddNodeTask(player, node, taskTypeInfo);

            SabotageImprovementEffect sabotage = new SabotageImprovementEffect();
            RepairImprovementEffect repair = new RepairImprovementEffect();
            sabotage.Initialize(Container);
            repair.Initialize(Container);

            worldManager.SetNodeImprovement(node, improvementType);
            Assert.IsFalse(node.Improvement.Sabotaged);

            sabotage.Fire(task);
            Assert.IsTrue(node.Improvement.Sabotaged);

            repair.Fire(task);
            Assert.IsFalse(node.Improvement.Sabotaged);
        }

        [TestMethod]
        public void NodeEffectTests_WorkImprovement()
        {
            //Worker types
            var generalType = Container.AddWorkerTypeInfo(WorkerTypeCodes.Generalist, 0);
            var specialistType = Container.AddWorkerTypeInfo(WorkerTypeCodes.Farmer, 1);
            
            //Improvement Type
            var improvementType = Container.AddImprovementTypeInfo(ImprovementTypeCodes.Farm, 1, 5);
            improvementType.ResourceProductionType = foodResourceType;
            improvementType.ResourceProductionRate = 5;
            improvementType.SpecialistWorkerType = specialistType;

            //Create Workers
            var generalWorker = playerManager.AddWorker(player, generalType);
            var specialistWorker = playerManager.AddWorker(player, specialistType);

            var taskTypeInfo = new NodeTaskTypeInfo(NodeTaskTypeCodes.WorkImprovement, null, null);
            var task = playerManager.AddNodeTask(player, node, taskTypeInfo);
            playerManager.AssignWorker(task, generalWorker);
            playerManager.AssignWorker(task, specialistWorker);

            worldManager.SetNodeImprovement(node, improvementType);

            //Create Work Improvement Effect
            WorkImprovementEffect workImprovementEffect = new WorkImprovementEffect();
            workImprovementEffect.Initialize(Container);
            
            //Call no action methods
            workImprovementEffect.Apply(task);
            workImprovementEffect.Refresh(task);
            workImprovementEffect.Remove(task);

            //Only one of relevance
            workImprovementEffect.Fire(task);

            Assert.AreEqual(15, player.Resources[foodResourceType]);
        }
    }
}
