﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Buildcraft.Core.Model;
using Buildcraft.Core.PlayerActions;
using Buildcraft.Core.Components;

using Buildcraft.Core.NodeTasks.Conditions;
using Rhino.Mocks;
using Buildcraft.ScenarioBuilder;

namespace Buildcraft.Core.Test.ModelActionTests
{
    [TestClass]
    public class TickWorldFireTasksTest : BaseTest
    {
        //Player Actions
        protected PlayerActions.PlayerActions _playerActions;

        //Tick Action
        protected TickWorldAction tickWorldAction;

        //World State
        private Player player;
        private Node node;


        private ResourceTypeInfo foodResourceType;
        private ResourceTypeInfo moneyResourceType;


        private const string LongConditionCode = "LongCondition";
        private const string LongEffectCode = "LongEffect";

        private ICondition longCondition = null;
        private IEffect<NodeTask> longEffect = null;

        private const string ShortConditionCode = "ShortCondition";
        private const string ShortEffectCode = "ShortEffect";

        private ICondition shortCondition = null;
        private IEffect<NodeTask> shortEffect = null;

        private ImprovementTypeInfo improvementType;
        private WorkerTypeInfo workerType;

        private NodeTaskTypeInfo longTaskType;
        private NodeTaskTypeInfo shortTaskType;


        protected override void CustomInitialize()
        {
            //Mock IEffect and ICondition
            longEffect = MockRepository.GenerateMock<IEffect<NodeTask>>();
            shortEffect = MockRepository.GenerateMock<IEffect<NodeTask>>();

            longCondition = MockRepository.GenerateMock<ICondition>();
            shortCondition = MockRepository.GenerateMock<ICondition>();

            Container.RegisterInstance(longEffect, LongEffectCode);
            Container.RegisterInstance(shortEffect, ShortEffectCode);

            Container.RegisterInstance(longCondition, LongConditionCode);
            Container.RegisterInstance(shortCondition, ShortConditionCode);
        }

        protected override void ConfigureWorld()
        {
            foodResourceType = new ResourceTypeInfo(ResourceTypeCodes.Food);
            moneyResourceType = new ResourceTypeInfo(ResourceTypeCodes.Money);

            //Create Improvement Type
            improvementType = Container.AddImprovementTypeInfo(ImprovementTypeCodes.Farm, 1, 5);

            //Create Worker Type
            workerType = Container.AddWorkerTypeInfo(WorkerTypeCodes.Generalist, 0);
            workerType.Consumption[foodResourceType] = 2;
            workerType.Consumption[moneyResourceType] = 2;

            //Create Long Task Type
            longTaskType = Container.AddNodeTaskTypeInfo(NodeTaskTypeCodes.TrainFarmer, LongConditionCode, LongEffectCode, 100, false);

            //Create Short Task Type
            shortTaskType = Container.AddNodeTaskTypeInfo(NodeTaskTypeCodes.TrainSoldier, ShortConditionCode, ShortEffectCode, 10, false);

            //Create Player And Node
            player = worldManager.AddPlayer("P1");
            node = CreateNode(player, improvementType);

            //Create Player Actions
            _playerActions = new PlayerActions.PlayerActions(player, Container);

            //Create Tick World Action
            tickWorldAction = Container.RetrieveInstance<TickWorldAction>();

            Container.Initialize();
        }

        [TestMethod]
        public void TickWorldTest_TaskFiresOnCompletion()
        {
            //Create Farm Task
            var workers = player.UnassignedWorkers.Where(w => w.TypeInfo == workerType).ToList();
            
            //Add Long Task
            var result = _playerActions.AddTask(node, longTaskType, workers);
            Assert.AreEqual(TaskValidityState.Ok, result);
            var longTask = node.Tasks[0];
            longCondition.AssertWasCalled(c => c.CheckNodeValidity(node, player));
            longEffect.AssertWasCalled(e => e.Apply(longTask));

            //Add Short Task
            result = _playerActions.AddTask(node, shortTaskType, workers);
            Assert.AreEqual(TaskValidityState.Ok, result);
            var shortTask = node.Tasks[1];
            shortCondition.AssertWasCalled(c => c.CheckNodeValidity(node, player));
            shortEffect.AssertWasCalled(e => e.Apply(shortTask));

            //Tick to complete short task
            tickWorldAction.TickWorld(10);
            shortEffect.AssertWasCalled(e => e.Fire(shortTask));
            shortEffect.AssertWasCalled(e => e.Remove(shortTask));
            Assert.IsTrue(node.Tasks.Count == 1);

            //Tick to complete long task
            tickWorldAction.TickWorld(90);
            longEffect.AssertWasCalled(e => e.Fire(longTask));
            longEffect.AssertWasCalled(e => e.Remove(longTask));
            Assert.IsTrue(node.Tasks.Count == 0);
        }

    }
}
