﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Core.PlayerActions;
using Buildcraft.Core.NodeTasks.Effects;
using Rhino.Mocks;
using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.ScenarioBuilder;

namespace Buildcraft.Core.Test.TypeInfoTests
{
    [TestClass]
    public class TaskTypeInfoTest : BaseTest
    {
        private Player player;
        private Player enemyPlayer;

        private Node node;
        private Node enemyNode;

        private ImprovementTypeInfo improvementType;
        private WorkerTypeInfo workerType;
        private NodeTaskTypeInfo taskType;
       

        private ICondition taskCondition;
        private IEffect<Model.NodeTask> taskEffect;

        protected override void CustomInitialize()
        {

        }

        protected override void ConfigureWorld()
        {
            player = worldManager.AddPlayer("P1");
            enemyPlayer = worldManager.AddPlayer("P2");

            node = worldManager.AddNode();
            worldManager.SetNodeOwner(node, player);
            worldManager.SetNodeImprovement(node, improvementType);

            enemyNode = worldManager.AddNode();
            worldManager.SetNodeOwner(enemyNode, enemyPlayer);

            improvementType = new ImprovementTypeInfo(ImprovementTypeCodes.Farm);
            improvementType.MinWorkers = 1;
            improvementType.MaxWorkers = 2;

            workerType = new WorkerTypeInfo(WorkerTypeCodes.Farmer);

            taskCondition = new WorkImprovementCondition();
            taskEffect = MockRepository.GenerateMock<IEffect<NodeTask>>();
            taskType = new NodeTaskTypeInfo(NodeTaskTypeCodes.WorkImprovement, "condition", "effect");

            Container.RegisterInstance<ICondition>(taskCondition, "condition");
            Container.RegisterInstance<IEffect<NodeTask>>(taskEffect, "effect");
            Container.RegisterInstance<NodeTaskTypeInfo>(taskType, NodeTaskTypeCodes.WorkImprovement);

            Container.Initialize();
        }

        [TestMethod]
        public void TaskTypeInfoTests_CheckNodeValidity_DuplicateTask()
        {
            var result = taskType.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.Ok, result);

            var task = playerManager.AddNodeTask(player, node, taskType);

            result = taskType.CheckNodeValidity(node, player);
            Assert.AreEqual(TaskValidityState.TaskAlreadyExistsOnNode, result);

            playerManager.RemoveNodeTask(task);
        }


        [TestMethod]
        public void TaskTypeInfoTests_CheckFailsTickWhenTaskInvalid()
        {
            var player = worldManager.AddPlayer("P1");
            var node = worldManager.AddNode();

            taskType.TaskDuration = 100;

            worldManager.SetNodeOwner(node, player);
            worldManager.SetNodeImprovement(node, improvementType);

            var worker = playerManager.AddWorker(player, workerType);
            var task = playerManager.AddNodeTask(player, node, taskType);

            //No Workers assigned

            var taskState = task.Type.Tick(task, 1);
            Assert.AreEqual(TaskState.Failed, taskState);
        }

        [TestMethod]
        public void TaskTypeInfoTests_CheckTaskCompletes()
        {
            var player = worldManager.AddPlayer("P1");
            var node = worldManager.AddNode();

            taskType.TaskDuration = 10;
            taskType.Recycle = false;

            worldManager.SetNodeOwner(node, player);
            worldManager.SetNodeImprovement(node, improvementType);

            var worker = playerManager.AddWorker(player, workerType);

            var task = playerManager.AddNodeTask(player, node, taskType);
            playerManager.AssignWorker(task, worker);

            var taskState = task.Type.Tick(task, 10);
            Assert.AreEqual(TaskState.Complete, taskState);
        }

        [TestMethod]
        public void TaskTypeInfoTests_CheckTaskRecycles()
        {
            taskType.TaskDuration = 10;
            taskType.Recycle = true;

            worldManager.SetNodeOwner(node, player);
            worldManager.SetNodeImprovement(node, improvementType);

            var worker = playerManager.AddWorker(player, workerType);

            var task = playerManager.AddNodeTask(player, node, taskType);
            playerManager.AssignWorker(task, worker);

            var taskState = taskType.Tick(task, 10);
            Assert.AreEqual(TaskState.Recycling, taskState);
        }

        [TestMethod]
        public void TaskTypeInfoTests_TaskApplyRemoveEffect()
        {
            var player = worldManager.AddPlayer("P1");
            var node = worldManager.AddNode();

            taskType.TaskDuration = 10;
            taskType.Recycle = true;
            //taskType.Effect = effect;

            worldManager.SetNodeOwner(node, player);
            worldManager.SetNodeImprovement(node, improvementType);

            var worker = playerManager.AddWorker(player, workerType);

            var task = playerManager.AddNodeTask(player, node, taskType);
            playerManager.AssignWorker(task, worker);


            task.ApplyEffects();
            taskEffect.AssertWasCalled(t => t.Apply(task));

            task.FireEffects();
            taskEffect.AssertWasCalled(t => t.Fire(task));

            task.RemoveEffects();
            taskEffect.AssertWasCalled(t => t.Remove(task));
        }
    }
}
