﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Core;
using Buildcraft.Core.Model.TypeInfo;
using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.Core.NodeTasks.Effects;
using Buildcraft.Core.PlayerTasks;
using Buildcraft.Core.PlayerTasks.Effects;

namespace Buildcraft.Core.Test
{
    public class WorldBuilder
    {
        //TODO: Refactor - Break down parts

        private IContainer _container;

        //World and Tasks
        private World world;

        private WorldManager worldManager;
        private NodeManager nodeManager;
        private PlayerManager playerManager;

        private NodeTaskManager nodeTaskManager;
        private PlayerTaskManager playerTaskManager;

        //Players
        List<Player> players = new List<Player>();

        //Player Capitals
        List<Node> capitals = new List<Node>();

        //Improvement Types
        protected ImprovementTypeInfo cityImprovement;
        protected ImprovementTypeInfo farmImprovement;

        //Worker Types
        protected WorkerTypeInfo generalWorker;
        protected WorkerTypeInfo farmWorker;
        protected WorkerTypeInfo soldierWorker;

        //Task Types
        protected NodeTaskTypeInfo trainFarmerTask;
        protected NodeTaskTypeInfo trainSoldierTask;
        protected NodeTaskTypeInfo workImprovementTask;
        protected NodeTaskTypeInfo repairImprovementTask;
        protected NodeTaskTypeInfo patrolBordersTask;
        protected NodeTaskTypeInfo attackNode;
        protected NodeTaskTypeInfo sabotageImprovement;

        //Player Tasks
        protected PlayerTaskTypeInfo workerConsumptionTask;
        protected PlayerTaskTypeInfo workerSpawnTask;

        public static class ScenarioCodes
        {
            public const string Empty = "Empty";
            public const string Line = "Line";
            public const string Square = "Square";
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="container"></param>
        /// <param name="scenarioCode"></param>
        /// <returns>A map of nodes to locations.</returns>
        public Dictionary<Node, MapPoint> BuildScenario(IContainer container, string scenarioCode)
        {
            if (container == null)
                throw new ArgumentNullException();

            this._container = container;

            //Unpack container and configure world
            UnpackContainer();
            InitializeWorld();

            int seed = 50;

            //Build Scenario
            switch (scenarioCode)
            {
                case ScenarioCodes.Empty:
                    return new Dictionary<Node, MapPoint>();

                case ScenarioCodes.Square:
                    return BuildSquareScenario(seed);

                case ScenarioCodes.Line:
                    return BuildLineScenario(seed);

                default:
                    throw new ApplicationException("Unhandled Scenario Code; " + scenarioCode);
            }

        }

        private void UnpackContainer()
        {
            world = _container.RetrieveInstance<World>();
            worldManager = _container.RetrieveInstance<WorldManager>();
            nodeManager = _container.RetrieveInstance<NodeManager>();
            nodeTaskManager = _container.RetrieveInstance<NodeTaskManager>();
            playerManager = _container.RetrieveInstance<PlayerManager>();
            playerTaskManager = _container.RetrieveInstance<PlayerTaskManager>();
        }


        private void InitializeWorld()
        {
            InitializeImprovementTypes();
            InitializeWorkerTypes();
            InitializeNodeConditions();
            InitializeNodeEffects();
            InitializePlayerEffects();
            InitializeNodeTaskTypes();
            InitializePlayerTaskTypes();

            _container.Initialize();
        }

        private void InitializeNodeConditions()
        {
            _container.RegisterInstance<ICondition>(
               new TrainWorkerCondition()
               {
                   MaxWorkers = 1,
                   MinWorkers = 1,
                   ImprovementRestriction = cityImprovement,
                   TargetWorkerType = farmWorker
               },
               NodeConditionCodes.TrainFarmerCondition);

            _container.RegisterInstance<ICondition>(
               new TrainWorkerCondition()
               {
                   MaxWorkers = 1,
                   MinWorkers = 1,
                   ImprovementRestriction = cityImprovement,
                   TargetWorkerType = soldierWorker
               },
               NodeConditionCodes.TrainSoldierCondition);

            _container.RegisterInstance<ICondition>(
                new RepairImprovementCondition()
                {
                    MinWorkers = 1,
                    MaxWorkers = 5
                },
                NodeConditionCodes.RepairImprovementCondition);

            _container.RegisterInstance<ICondition>(
                new PatrolBorderCondition()
                {
                    ImprovementRestriction = cityImprovement,
                    MinWorkers = 1,
                    MaxWorkers = 5
                },
                NodeConditionCodes.DefendTerritoryCondition);

            _container.RegisterInstance<ICondition>(
                 new AttackNodeCondition()
                 {
                     MinWorkers = 1,
                     MaxWorkers = 5
                 },
                 NodeConditionCodes.AttackNodeCondition);

            _container.RegisterInstance<ICondition>(
                new SabotageImprovementCondition()
                {
                    MinWorkers = 1,
                    MaxWorkers = 5
                },
                NodeConditionCodes.SabotageImprovementCondition);


            _container.RegisterInstance<ICondition>(
                new WorkImprovementCondition(),
                NodeConditionCodes.WorkImprovementCondition);
        }

        private void InitializeNodeEffects()
        {
            _container.RegisterInstance<IEffect<NodeTask>>(
                new ChangeWorkerTypeEffect(WorkerTypeCodes.Farmer),
                NodeEffectCodes.TrainFarmerEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new ChangeWorkerTypeEffect(WorkerTypeCodes.Soldier),
                NodeEffectCodes.TrainSoldierEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new RepairImprovementEffect(),
                NodeEffectCodes.RepairImprovementEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new DefendTerritoryEffect()
                {
                    DefencePerWorker = 2,
                    SpecialistType = soldierWorker,
                    SpecialistMultiplier = 2
                },
                NodeEffectCodes.DefendTerritoryEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new AttackNodeEffect()
                {
                    AttackPerWorker = 1,
                    SpecialistMultiplier = 2,
                    SpecialistType = soldierWorker
                },
                NodeEffectCodes.AttackNodeEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                  new SabotageImprovementEffect(),
                  NodeEffectCodes.SabotageImprovementEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                    new WorkImprovementEffect(),
                    NodeEffectCodes.WorkImprovementEffect);
        }

        private void InitializePlayerEffects()
        {
            var workerConsumptionEffect = new WorkerConsumptionEffect();
            _container.RegisterInstance<IEffect<PlayerTask>>(workerConsumptionEffect, PlayerEffectCodes.WorkerConsumptionEffects);

            var workerSpawnEffect = new WorkerSpawnEffect();
            workerSpawnEffect.SpawnType = generalWorker;
            _container.RegisterInstance<IEffect<PlayerTask>>(workerSpawnEffect, PlayerEffectCodes.WorkerSpawnEffect);
        }

        private void InitializeImprovementTypes()
        {
            cityImprovement = _container.AddImprovementTypeInfo(ImprovementTypeCodes.City, 1, 5);
            cityImprovement.ResourceProductionType = ResourceType.Money;
            cityImprovement.ResourceProductionRate = 20;
            cityImprovement.TaxRate = 0;

            farmImprovement = _container.AddImprovementTypeInfo(ImprovementTypeCodes.Farm, 1, 5);
            farmImprovement.SpecialistWorkerType = farmWorker;
            farmImprovement.ResourceProductionType = ResourceType.Food;
            farmImprovement.ResourceProductionRate = 20;
            farmImprovement.TaxRate = 5;
        }

        private void InitializeWorkerTypes()
        {
            //Worker Types
            generalWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Generalist, 0);
            generalWorker.Consumption[ResourceType.Food] = 2;
            generalWorker.Consumption[ResourceType.Money] = 1;

            farmWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Farmer, 2);
            farmWorker.Consumption[ResourceType.Food] = 2;
            farmWorker.Consumption[ResourceType.Money] = 2;

            soldierWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Soldier, 3);
            soldierWorker.Consumption[ResourceType.Food] = 2;
            soldierWorker.Consumption[ResourceType.Money] = 4;
        }

        private void InitializeNodeTaskTypes()
        {

            trainSoldierTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.TrainSoldier,
                NodeConditionCodes.TrainSoldierCondition,
                NodeEffectCodes.TrainSoldierEffect,
                100,
                false);

            trainFarmerTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.TrainFarmer,
                NodeConditionCodes.TrainFarmerCondition,
                NodeEffectCodes.TrainFarmerEffect,
                100,
                false);

            repairImprovementTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.RepairImprovement,
                NodeConditionCodes.RepairImprovementCondition,
                NodeEffectCodes.RepairImprovementEffect,
                100,
                false);

            patrolBordersTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.DefendTerritory,
                NodeConditionCodes.DefendTerritoryCondition,
                NodeEffectCodes.DefendTerritoryEffect,
                100,
                true);

            attackNode = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.AttackNode,
                NodeConditionCodes.AttackNodeCondition,
                NodeEffectCodes.AttackNodeEffect,
                100,
                false);

            sabotageImprovement = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.SabotageImprovement,
                NodeConditionCodes.SabotageImprovementCondition,
                NodeEffectCodes.SabotageImprovementEffect,
                100,
                false);

            workImprovementTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.WorkImprovement,
                NodeConditionCodes.WorkImprovementCondition,
                NodeEffectCodes.WorkImprovementEffect,
                100,
                true);
        }

        private void InitializePlayerTaskTypes()
        {
            workerConsumptionTask = _container.AddPlayerTaskTypeInfo(
                PlayerTaskTypeCodes.WorkerConsumption,
                PlayerEffectCodes.WorkerConsumptionEffects,
                150,
                true);

            workerSpawnTask = _container.AddPlayerTaskTypeInfo(
                PlayerTaskTypeCodes.WorkerSpawn,
                PlayerEffectCodes.WorkerSpawnEffect,
                200,
                true);
        }




        private Dictionary<Node, MapPoint> BuildLineScenario(int seed)
        {
            var rand = new Random(seed);

            InitializePlayers(2);

            var player1 = players[0];
            var player2 = players[1];

            Dictionary<Node, MapPoint> map = new Dictionary<Node, MapPoint>();

            //Create P1Node
            var p1Capital = CreateNode(player1, cityImprovement);

            map.Add(p1Capital, new MapPoint(1, 1));

            //Start with player 1 capital
            Node previous = p1Capital;
            Node current = null;

            //Create line of nodes
            for (int i = 0; i < 10; i++)
            {
                //Create Node and Join
                current = CreateNode(null, farmImprovement);
                CreateJoin(previous, current);

                //Add To Map
                map.Add(current, new MapPoint(i + 2, 1));

                //Switch
                previous = current;
                current = null;
            }

            var p2Capital = CreateNode(player2, cityImprovement);
            map.Add(p2Capital, new MapPoint(12, 1));

            //End with player 2 capital
            CreateJoin(previous, p2Capital);


            //Setup Economy
            player1.Resources[ResourceType.Food] = 100;
            player1.Resources[ResourceType.Money] = 100;

            player2.Resources[ResourceType.Food] = 50;
            player2.Resources[ResourceType.Money] = 50;

            //Retur Node Map
            return map;
        }

        private Dictionary<Node, MapPoint> BuildSquareScenario(int seed)
        {
            //TODO: Use Random Seed
            var rand = new Random(seed);

            //Create Players
            InitializePlayers(4);
            var player1 = players[0];
            var player2 = players[1];
            var player3 = players[2];
            var player4 = players[3];

            //Setup Resources
            player1.Resources[ResourceType.Food] = 200;
            player1.Resources[ResourceType.Money] = 200;

            player2.Resources[ResourceType.Food] = 150;
            player2.Resources[ResourceType.Money] = 150;

            player3.Resources[ResourceType.Food] = 100;
            player3.Resources[ResourceType.Money] = 100;

            player4.Resources[ResourceType.Food] = 50;
            player4.Resources[ResourceType.Money] = 50;

            //Create Capitals
            var p1Capital = CreateNode(player1, cityImprovement);
            var p2Capital = CreateNode(player2, cityImprovement);
            var p3Capital = CreateNode(player3, cityImprovement);
            var p4Capital = CreateNode(player4, cityImprovement);

            //Create Map Set Capitals
            Node[,] mapArray = new Node[5, 5];
            mapArray[0, 0] = p1Capital;
            mapArray[4, 4] = p2Capital;
            mapArray[0, 4] = p3Capital;
            mapArray[4, 0] = p4Capital;

            //Fill Map
            for (int x = 0; x < mapArray.GetLength(0); x++)
            {
                for (int y = 0; y < mapArray.GetLength(1); y++)
                {
                    //Skip pre populated nodes (capital)
                    if (mapArray[x, y] != null)
                        continue;

                    //Create Node and Join
                    mapArray[x, y] = CreateNode(null, farmImprovement);
                }
            }

            //Create Joins
            int[][] VectorMap = new int[][] { new int[] { 1, 0 }, new int[] { 1, 1 }, new int[] { 0, 1 }, new int[] { -1, 1 } };
            for (int x = 0; x < mapArray.GetLength(0); x++)
            {
                for (int y = 0; y < mapArray.GetLength(1); y++)
                {
                    var fromNode = mapArray[x, y];

                    foreach(var vector in VectorMap)
                    {
                        var joinToX = x + vector[0];
                        var joinToY = y + vector[1];

                        if (joinToX < 0 || joinToX > mapArray.GetLength(0) - 1)
                            continue;

                        if (joinToY < 0 || joinToY > mapArray.GetLength(1) - 1)
                            continue;

                        var toNode = mapArray[joinToX, joinToY];
                        CreateJoin(fromNode, toNode);
                    }
                }
            }

            //Generate Map Dictionary
            Dictionary<Node, MapPoint> map = new Dictionary<Node, MapPoint>();
            for (int x = 0; x < mapArray.GetLength(0); x++)
            {
                for (int y = 0; y < mapArray.GetLength(1); y++)
                {
                    map.Add(mapArray[x, y], new MapPoint(x + 1, y + 1));
                }
            }

            return map;
        }
        



        private void InitializePlayers(int playerCount)
        {
            //Create Players
            for (int i = 0; i < playerCount; i++)
            {
                players.Add(CreatePlayer("P" + i.ToString()));
            }
        }

        private Player CreatePlayer(string playerName)
        {
            //Create Player
            var player = worldManager.AddPlayer(playerName);

            //Add Default tasks, will possibly need to be extended.
            playerTaskManager.AddTask(player, workerSpawnTask);
            playerTaskManager.AddTask(player, workerConsumptionTask);

            return player;
        }

        private Node CreateNode(Player owner = null, ImprovementTypeInfo typeInfo = null)
        {
            //Create Node, assign ownership and set improvement
            var node = worldManager.AddNode();

            if (typeInfo != null)
                nodeManager.SetImprovement(node, typeInfo);

            if (owner != null)
                nodeManager.SetOwner(node, owner);

            return node;
        }

        private void CreateJoin(Node from, Node to)
        {
            worldManager.AddJoin(from, to);
        }
    }

    public class MapPoint
    {
        public MapPoint(int x, int y)
        {
            X = x;
            Y = y;
        }

        public int X;
        public int Y;
    }

    public static class NodeEffectCodes
    {
        public const string TrainFarmerEffect = "TrainFarmerEffect";
        public const string TrainSoldierEffect = "TrainSoldierEffect";

        public const string RepairImprovementEffect = "RepairImprovementEffect";
        public const string DefendTerritoryEffect = "DefendTerritoryEffect";
        public const string AttackNodeEffect = "AttackNodeEffect";
        public const string SabotageImprovementEffect = "SabotageImprovementEffect";
        public const string WorkImprovementEffect = "WorkImprovementEffect";
    }

    public static class NodeConditionCodes
    {
        public const string TrainSoldierCondition = "TrainSoldierCondition";
        public const string TrainFarmerCondition = "TrainFarmerCondition";

        public const string RepairImprovementCondition = "RepairImprovementCondition";
        public const string DefendTerritoryCondition = "DefendTerritoryCondition";
        public const string AttackNodeCondition = "AttackNodeCondition";
        public const string SabotageImprovementCondition = "SabotageImprovementCondition";
        public const string WorkImprovementCondition = "WorkImprovementCondition";
    }

    public static class PlayerEffectCodes
    {
        public const string WorkerConsumptionEffects = "WorkerConsumptionEffects";
        public const string WorkerSpawnEffect = "WorkerSpawnEffect";
    }

    public static class PlayerTaskTypeCodes
    {
        public const string WorkerConsumption = "WorkerConsumption";
        public const string WorkerSpawn = "WorkerSpawn";
    }
}
