﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Interfaces;
using Buildcraft.Core.Exceptions;
using Buildcraft.Core.UnitTasks;

namespace Buildcraft.Core.Components
{
    public class BasicTaskScheduler : ITaskScheduler
    {
        private List<ITaskProcessor> _processors = new List<ITaskProcessor>();
        
        private int _frameLength;
        private int _frameProgress;

        public BasicTaskScheduler(int frameLength)
        {
            if (frameLength <= 0)
                throw new BuildcraftException("frameLength must be greater than 0.");

            _frameLength = frameLength;
            _frameProgress = 0;
        }

        #region ITaskScheduler Members

        public void RegisterTaskProcessor(ITaskProcessor processor)
        {
            if (_processors.Contains(processor))
                throw new BuildcraftException("Processor already registered");

            _processors.Add(processor);
        }

        public void UnregisterTaskProcessor(ITaskProcessor processor)
        {
            if (!_processors.Remove(processor))
                throw new BuildcraftException("Processor not registered");
        }


        public void RegisterStateBucket(IUnitTaskStateBucket bucket)
        {
            foreach (var processor in _processors)
            {
                processor.RegisterStateBucket(bucket);
            }
        }

        public void UnregisterStateBucket(IUnitTaskStateBucket bucket)
        {
            foreach (var processor in _processors)
            {
                processor.UnregisterStateBucket(bucket);
            }
        }


        public void RegisterState(IUnitTaskStateItem state)
        {
            foreach (var processor in _processors)
            {
                processor.RegisterState(state);
            }
        }

        public void UnregisterState(IUnitTaskStateItem state)
        {
            foreach (var processor in _processors)
            {
                processor.UnregisterState(state);
            }
        }


        public void Tick(int ticks)
        {
            //Increment frame progress, and skip if we're still progressing through the frame.
            _frameProgress += ticks;
            if (_frameProgress < _frameLength)
                return;

            //Subtract the frame size, and then fire all processes.
            _frameProgress -= _frameLength;
            foreach (var processor in _processors)
            {
                processor.Process();
            }
        }

        #endregion
    }
}
