﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;

using Buildcraft.Core.Exceptions;
using Buildcraft.Core.Interfaces.UnitTasks.TaskInfo;
using Buildcraft.Core.Interfaces.UnitTasks;
using Buildcraft.Core.PlayerTasks;
using Buildcraft.Core.Interfaces;
using Buildcraft.Core.UnitTasks;

namespace Buildcraft.Core.Components
{
    public class PlayerManager : IInitializable
    {
        private World _world;

        private IHostNotify _hostNotify;
        private ITaskScheduler _taskScheduler;

        #region IInitializable Members

        public void Initialize(IBuildcraftContainer container)
        {
            _world = container.RetrieveInstance<World>();
            _hostNotify = container.RetrieveInstance<IHostNotify>();
            _taskScheduler = container.RetrieveInstance<ITaskScheduler>();
        }

        #endregion


        public PlayerTask AddPlayerTask(Player owner, PlayerTaskTypeInfo type)
        {
            //TODO: Guard Condition
            if (owner == null)
                throw new ApplicationException("owner");

            if (type == null)
                throw new ApplicationException("type");

            var task = new PlayerTask();
            task.TaskId = _world.GetNewtaskId();
            task.Owner = owner;
            task.Type = type;

            owner.PlayerTasks.Add(task);

            return task;
        }

        public void RemovePlayerTask(PlayerTask task)
        {
            if (!task.Owner.PlayerTasks.Remove(task))
                throw new BuildcraftException("Cannot remove task from player");
        }


        public NodeTask AddNodeTask(Player owner, Node node, NodeTaskTypeInfo type)
        {
            var task = new NodeTask();
            task.TaskId = _world.GetNewtaskId();
            task.Node = node;
            task.Owner = owner;
            task.Type = type;

            owner.NodeTasks.Add(task);
            node.Tasks.Add(task);

            _hostNotify.TaskAdded(task);

            return task;
        }

        public void RemoveNodeTask(NodeTask task)
        {
            _hostNotify.TaskRemoved(task);

            ClearWorkers(task);

            if (!task.Owner.NodeTasks.Remove(task))
                throw new BuildcraftException("Cannot remove task from player");

            if (!task.Node.Tasks.Remove(task))
                throw new BuildcraftException("Cannot remove task from node");
        }


        public Worker AddWorker(Player player, WorkerTypeInfo typeInfo)
        {
            if (player == null)
                throw new ArgumentNullException("player");

            if (typeInfo == null)
                throw new ArgumentNullException("typeInfo");

            var worker = new Worker();

            worker.Owner = player;
            worker.TypeInfo = typeInfo;
            worker.Task = null;
            worker.Id = _world.GetNewWorkerId();

            player.UnassignedWorkers.Add(worker);
            player.AllWorkers.Add(worker);

            return worker;
        }

        public void RemoveWorker(Worker worker)
        {
            if (worker.Task != null)
                throw new BuildcraftException("Worker currently assigned to task.");

            if (!worker.Owner.UnassignedWorkers.Remove(worker))
                throw new BuildcraftException("Unable to remove worker from UnassignedWorkers collection.");

            if (!worker.Owner.AllWorkers.Remove(worker))
                throw new BuildcraftException("Unable to remove worker from AllWorkers collection");
        }


        public void AssignWorker(NodeTask task, Worker worker)
        {
            if (task.Owner != worker.Owner)
                throw new BuildcraftException("Task and Worker owner does not match.");

            if (worker.Task != null)
                throw new BuildcraftException("Worker is already assigned.");

            task.Workers.Add(worker);
            worker.Task = task;

            if (!task.Owner.UnassignedWorkers.Remove(worker))
                throw new BuildcraftException("Unable to remove worker from unassigned pool.");
        }

        public void UnassignWorker(NodeTask task, Worker worker)
        {
            if (task.Owner != worker.Owner)
                throw new BuildcraftException("Task and Worker owner does not match.");

            if (!task.Workers.Remove(worker))
                throw new BuildcraftException("Unable to remove worker from unassigned pool.");

            worker.Task = null;
            task.Owner.UnassignedWorkers.Add(worker);
        }

        public void ClearWorkers(NodeTask task)
        {
            while (task.Workers.Count > 0)
            {
                UnassignWorker(task, task.Workers[task.Workers.Count - 1]);
            }
        }


        public Unit AddUnit(Player player, UnitTypeInfo typeInfo)
        {
            if (player == null)
                throw new ArgumentNullException("player");

            if (typeInfo == null)
                throw new ArgumentNullException("typeInfo");

            Unit unit = new Unit();
            unit.Id = _world.GetNewWorkerId();
            unit.Owner = player;
            unit.TypeInfo = typeInfo;
            unit.Node = null;

            player.AllUnits.Add(unit);
            player.UnassignedUnits.Add(unit);

            return unit;
        }

        public void RemoveUnit(Unit unit)
        {
            if (unit.TaskState != null)
                throw new BuildcraftException("Unit currently assigned to task.");

            if (!unit.Owner.UnassignedUnits.Remove(unit))
                throw new BuildcraftException("Unable to remove Units from UnassignedUnits collection.");

            if (!unit.Owner.AllUnits.Remove(unit))
                throw new BuildcraftException("Unable to remove unit from AllUnits collection");
        }


        public void AssignUnit(Node node, Unit unit, IUnitTaskInfo taskInfo)
        {
            if (unit.TaskState != null)
                throw new BuildcraftException("Worker is already assigned.");

            //Create task state for unit
            var taskState = taskInfo.CreateTaskState(node, unit);
            unit.TaskState = taskState;
            unit.Node = node;
            node.Units.Add(unit);

            taskState.Recalculate();
            
            //Get or create bucket, and assign task.
            IUnitTaskStateBucket bucket = GetBucket(node, unit.Owner, taskInfo);
            bucket.AddTaskState(taskState);

            //Assign to scheduler
            _taskScheduler.RegisterState(taskState);

            //Remove from unasigned units
            if (!unit.Owner.UnassignedUnits.Remove(unit))
                throw new BuildcraftException("Unable to remove worker from unassigned pool.");
        }

        public void UnassignUnit(Unit unit)
        {
            var owner = unit.Owner;
            var node = unit.Node;
            var bucket = GetExistingBucket(unit.Node, unit.Owner, unit.TaskState.TaskInfo);

            //Remove task state from bucket and unregister from scheduler
            bucket.RemoveTaskState(unit.TaskState);
            _taskScheduler.UnregisterState(unit.TaskState);

            //If bucket is empty, then remove the bucket from the node, otherwise, recalculate.
            if (bucket.TaskStates.Count == 0)
            {
                node.TaskStateBuckets.Remove(bucket);
                _taskScheduler.UnregisterStateBucket(bucket);
            }
            else
                bucket.Recalculate();

            //Unassign Unit and return to unasigned pool
            owner.UnassignedUnits.Add(unit);
            node.Units.Remove(unit);
            
            unit.Node = null;
            unit.TaskState = null;
        }



        private IUnitTaskStateBucket GetBucket(Node node, Player player, IUnitTaskInfo taskInfo)
        {
            var bucket = GetExistingBucket(node, player, taskInfo);

            if (bucket == null)
            {
                //Create Bucket and assign
                bucket = taskInfo.CreateTaskStateBucket(node, player);
                node.TaskStateBuckets.Add(bucket);
                _taskScheduler.RegisterStateBucket(bucket);
            }

            return bucket;
        }

        private IUnitTaskStateBucket GetExistingBucket(Node node, Player player, IUnitTaskInfo taskInfo)
        {
            return node.TaskStateBuckets
                .Where(ts => ts.Node == node && ts.Owner == player && ts.TaskInfo == taskInfo)
                .SingleOrDefault();
        }
    }
}
