﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.Exceptions;

using System.Diagnostics;
using Buildcraft.Core.NodeTasks.Conditions;

namespace Buildcraft.Core.Components
{
    /// <summary>
    /// Manager for Players, Nodes and Joins at the world level.
    /// </summary>
    public class WorldManager : IInitializable
    {
        public World _world;
        public IBuildcraftContainer _container;

        #region IInitializable Members

        public void Initialize(IBuildcraftContainer container)
        {
            _container = container;
            _world = container.RetrieveInstance<World>();
        }

        #endregion


        /// <summary>
        /// Add a player to the world
        /// </summary>
        /// <param name="world"></param>
        /// <param name="playerName"></param>
        /// <returns></returns>
        public Player AddPlayer(string playerName)
        {
            var player = new Player();

            player.Id = _world.GetNewPlayerId();
            player.Name = playerName;

            _world.Players.Add(player);

            return player;
        }

        /// <summary>
        /// Remove the player
        /// </summary>
        /// <param name="world"></param>
        /// <param name="player"></param>
        public void RemovePlayer(Player player)
        {
            if (!_world.Players.Remove(player))
                throw new BuildcraftException("Unable to remove player");
        }


        /// <summary>
        /// Add a node to the game world
        /// </summary>
        /// <param name="terrain"></param>
        /// <returns></returns>
        public Node AddNode()
        {
            Node node = new Node();
            node.Id = _world.GetNewNodeId();
            node.Owner = null;

            _world.Nodes.Add(node);

            return node;
        }

        /// <summary>
        /// Remove a node from the game world
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public void RemoveNode(Node node)
        {
            if (node.Owner != null)
                throw new BuildcraftException("Cannot remove node owned by player.");

            if (node.Tasks.Count > 0)
                throw new BuildcraftException("Cannot remove node with active tasks.");

            //Remove Joins from node
            foreach (var join in node.Joins)
            {
                RemoveJoin(join.FromNode, join.ToNode);
            }

            if (!_world.Nodes.Remove(node))
                throw new BuildcraftException("Unable to remove node");
        }


        /// <summary>
        /// Adds a bi-directional join between two nodes.
        /// </summary>
        /// <param name="from"></param>
        /// <param name="to"></param>
        public void AddJoin(Node from, Node to)
        {
            //Determine if join already exists
            if (from.Joins.Exists(j => j.FromNode == from && j.ToNode == to))
                throw new BuildcraftException("Join Already exists");

            int id = _world.GetNewJoinId();

            Join join1 = new Join();
            join1.JoinId = id;
            join1.FromNode = from;
            join1.ToNode = to;

            from.Joins.Add(join1);

            Join join2 = new Join();
            join2.JoinId = id;
            join2.FromNode = to;
            join2.ToNode = from;

            to.Joins.Add(join2);
        }

        /// <summary>
        /// Remove a join from betweeen two nodes.
        /// </summary>
        /// <param name="join"></param>
        public void RemoveJoin(Node from, Node to)
        {
            //Remove Outbound Joint
            var outJoin = from.Joins.Where(j => j.ToNode == to).Single();
            var inJoin = to.Joins.Where(j => j.ToNode == from).Single();

            if (!from.Joins.Remove(outJoin))
                throw new BuildcraftException("Unable to remove join from source node.");
            
            if (!to.Joins.Remove(inJoin))
                throw new BuildcraftException("Unable to remove join from target node.");
        }


        /// <summary>
        /// Set an improvement on a node.
        /// </summary>
        /// <param name="node">The node to set the improvement on.</param>
        /// <param name="typeInfo">The improvement to create</param>
        public void SetNodeImprovement(Node node, ImprovementTypeInfo typeInfo)
        {
            ClearNodeImprovement(node);

            node.Improvement = new Improvement() { TypeInfo = typeInfo, ImprovementLevel = 0 };
        }

        /// <summary>
        /// Clear an improvement from a node.
        /// </summary>
        /// <param name="node">The node to clear the improvement on.</param>
        public void ClearNodeImprovement(Node node)
        {
            node.Improvement = null;
        }


        /// <summary>
        /// Set the node owner.
        /// </summary>
        /// <param name="node">The node to set the owner on.</param>
        /// <param name="player">The player to set as the owner.</param>
        public void SetNodeOwner(Node node, Player player)
        {
            if (node.Owner != null)
                ClearNodeOwner(node);

            node.Owner = player;
            player.Nodes.Add(node);
        }

        /// <summary>
        /// Clear the node owner.
        /// </summary>
        /// <param name="node">The node to clear.</param>
        public void ClearNodeOwner(Node node)
        {
            if (node.Owner == null)
                throw new BuildcraftException("Node already clear");

            if (!node.Owner.Nodes.Remove(node))
                throw new BuildcraftException("Unable to remove node");

            node.Owner = null;
        }
    }
}
