﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.Core.Model;
using Buildcraft.Core.PlayerTasks;
using Buildcraft.Core.Interfaces.UnitTasks.TaskInfo;
using Buildcraft.Core.Interfaces.UnitTasks;
using Buildcraft.Core.UnitTasks;

namespace Buildcraft.Core
{
    /// <summary>
    /// Dependency injection container for the game
    /// </summary>
    public interface IBuildcraftContainer
    {
        /// <summary>
        /// Register an instance of a class
        /// </summary>
        /// <typeparam name="T">The Type to register.</typeparam>
        /// <param name="instance">The instance to register.</param>
        void RegisterInstance<T>(T instance);

        /// <summary>
        /// Register an instance of a class
        /// </summary>
        /// <typeparam name="T">The Type to register.</typeparam>
        /// <param name="instance">The instance to register.</param>
        /// <param name="objectCode">The code to registeragainst</param>
        void RegisterInstance<T>(T instance, string objectCode);


        /// <summary>
        /// Retrieve an instance of a class. Should throw exception if the instance is not found.
        /// </summary>
        /// <typeparam name="T">The Type to retrieve.</typeparam>
        /// <returns>The instance to retrieve.</returns>
        T RetrieveInstance<T>();

        /// <summary>
        /// Retrieve an instance of an item by code.
        /// </summary>
        /// <typeparam name="T">The type to retrieve</typeparam>
        /// <param name="objectCode">The code to retrieve against</param>
        /// <returns></returns>
        T RetrieveInstance<T>(string objectCode);

        /// <summary>
        /// Retrieve all instances in the container, registered against a given type.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <returns></returns>
        List<T> RetrieveAllInstances<T>();

        /// <summary>
        /// Initialize any instances in the container that implement IInitializable.
        /// </summary>
        void Initialize();
    }

    public static class IBuildcraftContainerExtensions
    {
        /// <summary>
        /// Add ImprovementTypeInfo
        /// </summary>
        /// <returns>The improvement type created.</returns>
        public static ImprovementTypeInfo AddImprovementTypeInfo(this IBuildcraftContainer container, string improvementTypeCode, int minWorkers, int maxWorkers)
        {
            //Guards
            if (minWorkers <= 0) throw new ArgumentException("Min workers must be greater than 0");
            if (maxWorkers < minWorkers) throw new ArgumentException("Max workers cannot be less than min");

            var typeInfo = new ImprovementTypeInfo(improvementTypeCode);
            typeInfo.MinWorkers = minWorkers;
            typeInfo.MaxWorkers = maxWorkers;

            container.RegisterInstance<ImprovementTypeInfo>(typeInfo, improvementTypeCode);

            return typeInfo;
        }

        /// <summary>
        /// Add a WorkerTypeInfo
        /// </summary>
        /// <returns>The worker type info created</returns>
        public static WorkerTypeInfo AddWorkerTypeInfo(this IBuildcraftContainer container, string workerTypeCode, int rank)
        {
            //TODO: Guards

            WorkerTypeInfo typeInfo = new WorkerTypeInfo(workerTypeCode);
            typeInfo.Rank = rank;

            container.RegisterInstance<WorkerTypeInfo>(typeInfo, workerTypeCode);

            return typeInfo;
        }

        /// <summary>
        /// Add TaskTypeInfo
        /// </summary>
        /// <returns>The instance created.</returns>
        public static NodeTaskTypeInfo AddNodeTaskTypeInfo(this IBuildcraftContainer container, string nodeTaskTypeCode, string conditionCode, string effectCode, int duration, bool recycle)
        {
            //TODO: Guards
            NodeTaskTypeInfo typeInfo = new NodeTaskTypeInfo(nodeTaskTypeCode, conditionCode, effectCode);
            typeInfo.TaskDuration = duration;
            typeInfo.Recycle = recycle;

            container.RegisterInstance<NodeTaskTypeInfo>(typeInfo, nodeTaskTypeCode);

            return typeInfo;
        }

        public static PlayerTaskTypeInfo AddPlayerTaskTypeInfo(this IBuildcraftContainer container, string playerTaskTypeCode, string effectCode, int duration, bool recycle)
        {
            PlayerTaskTypeInfo typeInfo = new PlayerTaskTypeInfo(playerTaskTypeCode, effectCode);
            typeInfo.TaskDuration = duration;
            typeInfo.Recycle = recycle;

            container.RegisterInstance<PlayerTaskTypeInfo>(typeInfo, playerTaskTypeCode);

            return typeInfo;
        }

        public static ResourceTypeInfo AddResourceTypeInfo(this IBuildcraftContainer container, string resourceTypeCode)
        {
            ResourceTypeInfo typeInfo = new ResourceTypeInfo(resourceTypeCode);

            container.RegisterInstance<ResourceTypeInfo>(typeInfo, resourceTypeCode);

            return typeInfo;
        }

        public static UnitTypeInfo AddUnitTypeInfo(this IBuildcraftContainer container, string unitCode)
        {
            var typeInfo = new UnitTypeInfo(unitCode);

            container.RegisterInstance<UnitTypeInfo>(typeInfo, unitCode);

            return typeInfo;
        }

        public static NodeEffectTypeInfo AddNodeEffectTypeInfo(this IBuildcraftContainer container, string effectCode)
        {
            var typeInfo = new NodeEffectTypeInfo(effectCode);

            container.RegisterInstance<NodeEffectTypeInfo>(typeInfo, effectCode);

            return typeInfo;
        }

        public static IUnitTaskInfo AddUnitTaskInfo(this IBuildcraftContainer container, string taskCode)
        {
            //TODO: Branch Between Type
            throw new NotImplementedException();
        }
    }
}
