﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.Exceptions;
using Buildcraft.Core.Interfaces;


namespace Buildcraft.Core.Components
{
    public class TickWorldAction : IInitializable
    {
        private World _world;

        private PlayerManager _playerManager;

        private IHostNotify _hostNotify;
        private ITaskScheduler _taskScheduler;
        
        public ResourceTypeInfo VictoryResourceType;

        List<NodeTask> _expiredNodeTasks = new List<NodeTask>();
        List<PlayerTask> _expiredPlayerTasks = new List<PlayerTask>();


        #region IInitializable Members

        public void Initialize(IBuildcraftContainer container)
        {
            _world = container.RetrieveInstance<World>();
            _playerManager = container.RetrieveInstance<PlayerManager>();
            _hostNotify = container.RetrieveInstance<IHostNotify>();
            _taskScheduler = container.RetrieveInstance<ITaskScheduler>();
        }

        #endregion


        public void TickWorld(int duration)
        {
            //World State Guard
            if (_world.GameState != GameState.InProgress)
                return;

            _expiredPlayerTasks.Clear();
            _expiredNodeTasks.Clear();

            TickPlayers(duration);
            TickNodes(duration);

            //NOTE: This is going here temporarily as it doesn't really have a well defined place at the moment. 
            _taskScheduler.Tick(duration);

            DetermineVictor();
        }

        private void DetermineVictor()
        {
            Player victor = null;

            foreach (var player in _world.Players)
            {
                //10 Nodes and 1000 money
                if (player.Nodes.Count >= 10 &&
                    player.Resources[VictoryResourceType] >= 1000)
                {
                    victor = player;
                    break;
                }

            }

            if (victor != null)
            {
                _world.GameState = GameState.GameOver;
                _world.Victor = victor;
                _hostNotify.GameOver(_world.Victor);
            }
        }

        private void TickPlayers(int duration)
        {
            //Enumerte all nodes, and update all tasks
            foreach (var player in _world.Players)
            {
                foreach (var task in player.PlayerTasks)
                {
                    var taskState = task.Tick(duration);

                    if (taskState == TaskState.Complete ||
                        taskState == TaskState.Recycling)
                        task.FireEffects();

                    if (taskState == TaskState.Recycling)
                        task.Progress = 0;

                    if (taskState == TaskState.Complete ||
                        taskState == TaskState.Failed)
                    {
                        _expiredPlayerTasks.Add(task);
                    }
                }
            }

            //Remove Redundant Tasks
            foreach (var task in _expiredPlayerTasks)
            {
                _playerManager.RemovePlayerTask(task);
            }
        }

        private void TickNodes(int duration)
        {
            //Enumerte all nodes, and update all tasks
            foreach (var node in _world.Nodes)
            {
                foreach (var task in node.Tasks)
                {
                    var taskState = task.Tick(duration);

                    if (taskState == TaskState.Complete ||
                        taskState == TaskState.Recycling)
                        task.FireEffects();

                    if (taskState == TaskState.Recycling)
                        task.Progress = 0;

                    if (taskState == TaskState.Complete ||
                        taskState == TaskState.Failed)
                    {
                        _expiredNodeTasks.Add(task);
                    }
                }
            }

            //Remove Redundant Tasks
            foreach (var task in _expiredNodeTasks)
            {
                task.RemoveEffects();
                _playerManager.RemoveNodeTask(task);
            }
        }



    }
}
