﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Exceptions;
using Buildcraft.Core.PlayerActions;
using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.Core.Interfaces;

namespace Buildcraft.Core.Model
{
    public class NodeTaskTypeInfo: IInitializable
    {
        public string TaskTypeCode;

        //Task Details
        public int TaskDuration;
        public bool Recycle;
        public string ConditionCode;
        public string EffectCode;
        
        //Task Structure
        private ICondition Condition;
        private IEffect<Model.NodeTask> Effect;
        private IHostNotify HostNotify;


        public NodeTaskTypeInfo(string typeCode, string conditionCode, string effectCode)
        {
            TaskTypeCode = typeCode;
            ConditionCode = conditionCode;
            EffectCode = effectCode;
        }

        public void Initialize(IBuildcraftContainer container)
        {
            Condition = container.RetrieveInstance<ICondition>(ConditionCode);
            Effect = container.RetrieveInstance<IEffect<NodeTask>>(EffectCode);
            HostNotify = container.RetrieveInstance<IHostNotify>();
        }


        public TaskValidityState CheckExistingTaskValidity(NodeTask task)
        {
            return CheckTaskValidity(task.Node, task.Owner, task.Workers, true);
        }

        public TaskValidityState CheckNewTaskValidity(Node node, Player player, IEnumerable<Worker> workers)
        {
            return CheckTaskValidity(node, player, workers, false);
        }


        public TaskValidityState CheckTaskValidity(Node node, Player player, IEnumerable<Worker> workers, bool existing)
        {
            TaskValidityState result;

            //Valid for node check
            if (!existing)
            {
                result = CheckNodeValidity(node, player);
                if (result != TaskValidityState.Ok)
                    return result;
            }

            //Valid for workers check
            return CheckWorkersValidity(node, player, workers);
        }


        public TaskValidityState CheckNodeValidity(Node node, Player player)
        {
            //Check no duplicate task exists.
            if (node.Tasks.Exists(t => t.Type == this && t.Owner == player))
                return TaskValidityState.TaskAlreadyExistsOnNode;

            return Condition.CheckNodeValidity(node, player);
        }

        public TaskValidityState CheckWorkerValidity(Node node, Player player, Worker worker)
        {
            return Condition.CheckWorkerValidity(node, player, worker);
        }

        public TaskValidityState CheckWorkersValidity(Node node, Player player, IEnumerable<Worker> workers)
        {
            return Condition.CheckWorkersValidity(node, player, workers);
        }


        public TaskState Tick(NodeTask task, int progress)
        {
            //Check task is valid
            if (CheckExistingTaskValidity(task) != TaskValidityState.Ok)
                return TaskState.Failed;

            //Progress Task
            task.Progress += progress;

            bool complete = (task.Progress >= this.TaskDuration);

            TaskState result;

            //Check for completeness
            if (!complete)
                result = TaskState.Working;
            else
            {
                result = this.Recycle ?
                    TaskState.Recycling :
                    TaskState.Complete;
            }

            return result;
        }


        public void ApplyEffects(NodeTask task)
        {
            if (CheckExistingTaskValidity(task) != TaskValidityState.Ok)
                throw new BuildcraftException("Cannot apply effect for invalid task.");

            Effect.Apply(task);
        }

        public void RefreshEffects(NodeTask task)
        {
            if (CheckExistingTaskValidity(task) != TaskValidityState.Ok)
                throw new BuildcraftException("Cannot refresh effect for invalid task.");

            Effect.Refresh(task);
        }

        public void FireEffects(NodeTask task)
        {
            if (CheckExistingTaskValidity(task) != TaskValidityState.Ok)
                throw new BuildcraftException("Cannot fire effect for invalid task.");

            Effect.Fire(task);
        }

        public void RemoveEffects(NodeTask task)
        {
            Effect.Remove(task);
        }


        public override string ToString()
        {
            return TaskTypeCode.ToString();
        }
    }
}
