﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Core.Exceptions;
using Buildcraft.Core.Interfaces.UnitTasks;
using Buildcraft.Core.UnitTasks;

namespace Buildcraft.Core.PlayerActions
{
    public class PlayerActions
    {
        private Player _player;
        private PlayerManager _playerManager;
        private IBuildcraftContainer _container;

        public PlayerActions(Player player, IBuildcraftContainer container)
        {
            _player = player;
            _playerManager = container.RetrieveInstance<PlayerManager>();
            _container = container;
        }


        public TaskValidityState CanAddTask(Node node, NodeTaskTypeInfo taskType, IEnumerable<Worker> workers)
        {
            return taskType.CheckNewTaskValidity(node, _player, workers);
        }

        public TaskValidityState AddTask(Node node, NodeTaskTypeInfo taskType, IEnumerable<Worker> workers)
        {
            //Guard Condition
            var checkState = taskType.CheckNewTaskValidity(node, _player, workers);
            if (checkState != TaskValidityState.Ok)
                return checkState;

            //Add Task
            var task = _playerManager.AddNodeTask(_player, node, taskType);

            //Populate The Task
            foreach (var worker in workers)
            {
                _playerManager.AssignWorker(task, worker);
            }

            //Apply Effects
            task.ApplyEffects();
            
            #if DEBUG
            //Confirm validity in debug
            if (taskType.CheckExistingTaskValidity(task) != TaskValidityState.Ok)
                throw new BuildcraftException("Task added in invalid state.");
            #endif

            return TaskValidityState.Ok;
        }


        public TaskValidityState CancelTask(NodeTask task)
        {
            _playerManager.RemoveNodeTask(task);
            return TaskValidityState.Ok;
        }


        public void GetTasksForNode(Node node, List<NodeTaskTypeInfo> taskList)
        {
            taskList.Clear();

            //Get tasks that can be added to the node, then add them to the internal collection
            var tasks = _container.RetrieveAllInstances<NodeTaskTypeInfo>().Where(t => t.CheckNodeValidity(node, _player) == TaskValidityState.Ok);
            taskList.AddRange(tasks);
        }


        public TaskValidityState CanModifyTask(NodeTask task, IEnumerable<Worker> workers)
        {
            return task.Type.CheckTaskValidity(task.Node, _player, workers, true);
        }

        public TaskValidityState ModifyTask(NodeTask task, IEnumerable<Worker> workers)
        {
            var taskType = task.Type;

            //Guard Condition
            var checkState = taskType.CheckTaskValidity(task.Node, _player, workers, true);
            if (checkState != TaskValidityState.Ok)
                return checkState;

            _playerManager.ClearWorkers(task);

            //Populate The Task
            foreach (var worker in workers)
            {
                _playerManager.AssignWorker(task, worker);
            }

#if DEBUG
            //Confirm validity in debug
            if (taskType.CheckExistingTaskValidity(task) != TaskValidityState.Ok)
                throw new BuildcraftException("Task added in invalid state.");
#endif

            return TaskValidityState.Ok;
        }



        public bool CanAssignUnitToTask(Node node, Unit unit, IUnitTaskInfo unitTask)
        {
            if (!ConstraintValidator.Validate(node, unit, unitTask))
                return false;

            //TODO: Test Capacity? Test bucket state? Anything else?

            return true;
        }

        public bool AssignUnitToTask(Node node, Unit unit, IUnitTaskInfo unitTask)
        {
            if (!CanAssignUnitToTask(node, unit, unitTask))
                return false;

            _playerManager.AssignUnit(node, unit, unitTask);
            return true;
        }


        public bool CanUnassignUnitFromTask(Unit unit)
        {
            //TODO: Test constraints

            throw new NotImplementedException();
        }

        public bool UnassignUnitFromTask(Unit unit)
        {
            if (!CanUnassignUnitFromTask(unit))
                return false;

            _playerManager.UnassignUnit(unit);
            return true;
        }
    }
}
