﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Interfaces;
using Buildcraft.Core.UnitTasks;
using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Core.Exceptions;

namespace Buildcraft.Core.TaskProcessors
{
    public class CombatProcessor : ITaskProcessor, IInitializable
    {
        private Dictionary<Player, List<IUnitTaskStateBucketRegionDefence>> _regionalDefenceGroupings = 
                new Dictionary<Player, List<IUnitTaskStateBucketRegionDefence>>();

        private List<Node> _combatNodes = new List<Node>();
        
        private WorldManager _worldManager;


        #region IInitializable Members

        public void Initialize(IBuildcraftContainer container)
        {
            _worldManager = container.RetrieveInstance<WorldManager>();
        }

        #endregion


        #region ITaskProcessor Members

        public void RegisterStateBucket(UnitTasks.IUnitTaskStateBucket bucket)
        {
            if (bucket is IUnitTaskStateBucketNodeDefence)
                RegisterNodeDefenceStateBucket((IUnitTaskStateBucketNodeDefence)bucket);

            //Add region defence buckets 
            if (bucket is IUnitTaskStateBucketRegionDefence)
                RegisterRegionDefenceStateBucket((IUnitTaskStateBucketRegionDefence)bucket);

            if (bucket is IUnitTaskStateBucketNodeAttack)
                RegisterNodeAttackStateBucket((IUnitTaskStateBucketNodeAttack)bucket);
        }

        public void UnregisterStateBucket(UnitTasks.IUnitTaskStateBucket bucket)
        {
            if (bucket is IUnitTaskStateBucketNodeDefence)
                UnregisterNodeDefenceStateBucket((IUnitTaskStateBucketNodeDefence)bucket);

            //Add region defence buckets 
            if (bucket is IUnitTaskStateBucketRegionDefence)
                UnregisterRegionDefenceStateBucket((IUnitTaskStateBucketRegionDefence)bucket);

            if (bucket is IUnitTaskStateBucketNodeAttack)
                UnregisterNodeAttackStateBucket((IUnitTaskStateBucketNodeAttack)bucket);
        }


        public void RegisterState(UnitTasks.IUnitTaskStateItem state)
        {
            throw new NotImplementedException();
        }

        public void UnregisterState(UnitTasks.IUnitTaskStateItem state)
        {
            throw new NotImplementedException();
        }


        public void Process()
        {
            foreach (var node in _combatNodes)
            {
                //Calculate Combat Values
                float attackPower = CalculateAttackPower(node);
                float defencePower = CalculateDefencePower(node);

                if (attackPower > defencePower)
                {
                    var primaryAttacker = DeterminePrimaryAttacker(node);

                    if (primaryAttacker == null)
                        throw new BuildcraftException("No primary attacker.");

                    _worldManager.SetNodeOwner(node, primaryAttacker);
                    
                    //TODO: Notify of node conquest
                }
                else
                {
                    //TODO: Notify node defended
                }
            }
        }

        #endregion


        private void RegisterNodeDefenceStateBucket(IUnitTaskStateBucketNodeDefence bucket)
        {
            //Do not store these - we can pull this information off the nodes when required.
        }

        private void RegisterRegionDefenceStateBucket(IUnitTaskStateBucketRegionDefence bucket)
        {
            //Add Mapping to defence dictionary
            if (!_regionalDefenceGroupings.ContainsKey(bucket.Owner))
                _regionalDefenceGroupings.Add(bucket.Owner, new List<IUnitTaskStateBucketRegionDefence>());

            //Add Bucket to groupings
            _regionalDefenceGroupings[bucket.Owner].Add((IUnitTaskStateBucketRegionDefence)bucket);
        }

        private void RegisterNodeAttackStateBucket(IUnitTaskStateBucketNodeAttack bucket)
        {
            if (!_combatNodes.Contains(bucket.Node))
                _combatNodes.Add(bucket.Node);
        }


        private void UnregisterNodeDefenceStateBucket(IUnitTaskStateBucketNodeDefence bucket)
        {
            //Do not store these - we can pull this information off the nodes when required.
        }

        private void UnregisterRegionDefenceStateBucket(IUnitTaskStateBucketRegionDefence bucket)
        {
            //Add Mapping to defence dictionary
            if (!_regionalDefenceGroupings.ContainsKey(bucket.Owner))
                return;

            //Add Bucket to groupings
            _regionalDefenceGroupings[bucket.Owner].Remove(bucket);
        }

        private void UnregisterNodeAttackStateBucket(IUnitTaskStateBucketNodeAttack bucket)
        {
            throw new NotImplementedException();
        }


        private float CalculateAttackPower(Node node)
        {
            float value = 0;
            
            //Get Total Attack states
            foreach (var bucket in node.TaskStateBuckets)
            {
                if (bucket is IUnitTaskStateBucketNodeAttack)
                {
                    var typedBucket = (IUnitTaskStateBucketNodeAttack)bucket;
                    value += typedBucket.Power;
                }
            }

            return value;
        }

        private float CalculateDefencePower(Node node)
        {
            float value = 0;

            //Get Total Node Defence
            foreach (var bucket in node.TaskStateBuckets)
            {
                if (bucket is IUnitTaskStateBucketNodeDefence)
                {
                    var typedBucket = (IUnitTaskStateBucketNodeDefence)bucket;
                    value += typedBucket.Power;
                }
            }

            //Add Player regional defence
            if (_regionalDefenceGroupings.ContainsKey(node.Owner))
                value += _regionalDefenceGroupings[node.Owner].Sum(b => b.Power);
            
            return value;
        }


        private Player DeterminePrimaryAttacker(Node node)
        {
            IUnitTaskStateBucketNodeAttack currentPrimary = null;
            
            foreach (var bucket in node.TaskStateBuckets)
            {
                var typedBucket = bucket as IUnitTaskStateBucketNodeAttack;

                if (typedBucket == null)
                    continue;

                if (currentPrimary == null)
                    currentPrimary = typedBucket;

                if (typedBucket.Power > currentPrimary.Power)
                    currentPrimary = typedBucket;
            }

            return currentPrimary == null ? null : currentPrimary.Owner;
        }
    }
}
