﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.Exceptions;

namespace Buildcraft.Core.UnitTasks
{
    /// <summary>
    /// This interface is implemented by any class that will be used to store data at a node level.
    /// </summary>
    public interface IUnitTaskStateBucket
    {
        /// <summary>
        /// The TaskInfo that this task state belongs too.
        /// </summary>
        IUnitTaskInfo TaskInfo { get; }


        /// <summary>
        /// The player that owns this task state bucket.
        /// </summary>
        Player Owner { get; }

        /// <summary>
        /// The node this task state is running on.
        /// </summary>
        Node Node { get; }


        /// <summary>
        /// The task states that are in this bucket
        /// </summary>
        List<IUnitTaskStateItem> TaskStates { get; }


        /// <summary>
        /// Add a task state to the bucket.
        /// </summary>
        /// <param name="taskState"></param>
        void AddTaskState(IUnitTaskStateItem taskState, bool recalc = true);

        /// <summary>
        /// Remove a task state from the bucket
        /// </summary>
        /// <param name="taskState"></param>
        void RemoveTaskState(IUnitTaskStateItem taskState, bool recalc = true);


        /// <summary>
        /// Recalculate the task state bucket.
        /// </summary>
        /// <param name="fullRecalc">If set to true, will recalc all task states in the bucket. If false, simply recalculates the bucket from existing values.</param>
        void Recalculate(bool fullRecalc = false);
    }

    public abstract class UnitTaskStateBucketBase : IUnitTaskStateBucket
    {
        protected IUnitTaskInfo _taskInfo;
        protected Node _node;
        protected Player _owner;
        protected List<IUnitTaskStateItem> _taskStates;

        protected UnitTaskStateBucketBase(IUnitTaskInfo taskInfo, Node node, Player owner)
        {
            _taskInfo = taskInfo;
            _node = node;
            _owner = owner;

            _taskStates = new List<IUnitTaskStateItem>();
        }


        /// <summary>
        /// The task info that this state bucket belongs to.
        /// </summary>
        public IUnitTaskInfo TaskInfo 
        {
            get { return _taskInfo; } 
        }


        /// <summary>
        /// The Node this task state bucket belongs to.
        /// </summary>
        public Node Node
        {
            get { return _node; }
        }

        /// <summary>
        /// The player that owns this task state bucket.
        /// </summary>
        public Player Owner
        {
            get { return _owner; }
        }

        /// <summary>
        /// The task states that are in this bucket
        /// </summary>
        public List<IUnitTaskStateItem> TaskStates
        {
            get { return _taskStates; }
        }


        /// <summary>
        /// Add a task state to the bucket.
        /// </summary>
        /// <param name="taskState"></param>
        public void AddTaskState(IUnitTaskStateItem taskState, bool recalc = true)
        {
            if (_taskStates.Contains(taskState))
                throw new BuildcraftException("State already in bucket");

            _taskStates.Add(taskState);

            if (recalc)
                Recalculate();
        }

        /// <summary>
        /// Remove a task state from the bucket
        /// </summary>
        /// <param name="taskState"></param>
        public void RemoveTaskState(IUnitTaskStateItem taskState, bool recalc = true)
        {
            if (!_taskStates.Remove(taskState))
                throw new BuildcraftException("State is not in bucket");

            if (recalc)
                Recalculate();
        }


        /// <summary>
        /// Recalculate the task state bucket.
        /// </summary>
        /// <param name="fullRecalc">If set to true, will recalc all task states in the bucket. If false, simply recalculates the bucket from existing values.</param>
        public void Recalculate(bool fullRecalc = false)
        {
            if (fullRecalc)
            {
                foreach (var state in _taskStates)
                {
                    state.Recalculate();
                }
            }

            InternalRecalculate();
        }


        /// <summary>
        /// Abstract internal method for recalculating the task state bucket.
        /// </summary>
        protected abstract void InternalRecalculate();
    }
}
