﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core;

using Buildcraft.Core.PlayerTasks;
using Buildcraft.Core.Components;

namespace Buildcraft.ScenarioBuilder
{
    public class ScenarioAssembler
    {
        private IBuildcraftContainer _container;

        //World and Tasks
        private World world;

        private WorldManager worldManager;
        private PlayerManager playerManager;

        //Players
        protected List<Player> players = new List<Player>();

        //Player Capitals
        protected List<Node> capitals = new List<Node>();

        //Resource Types
        protected ResourceTypeInfo foodResource;
        protected ResourceTypeInfo moneyResource;

        //Improvement Types
        protected ImprovementTypeInfo cityImprovement;
        protected ImprovementTypeInfo farmImprovement;

        //Worker Types
        protected WorkerTypeInfo generalWorker;
        protected WorkerTypeInfo farmWorker;
        protected WorkerTypeInfo soldierWorker;
        protected WorkerTypeInfo adminWorker;

        //Task Types
        protected NodeTaskTypeInfo trainFarmerTask;
        protected NodeTaskTypeInfo trainSoldierTask;
        protected NodeTaskTypeInfo workImprovementTask;
        protected NodeTaskTypeInfo repairImprovementTask;
        protected NodeTaskTypeInfo patrolBordersTask;
        protected NodeTaskTypeInfo attackNode;
        protected NodeTaskTypeInfo sabotageImprovement;

        //Player Tasks
        protected PlayerTaskTypeInfo workerConsumptionTask;
        protected PlayerTaskTypeInfo workerSpawnTask;


        /// <summary>
        /// 
        /// </summary>
        /// <param name="container"></param>
        /// <param name="scenarioCode"></param>
        /// <returns>A map of nodes to locations.</returns>
        public Dictionary<Node, MapPoint> BuildScenario(IBuildcraftContainer container, string scenarioCode)
        {
            if (container == null)
                throw new ArgumentNullException();

            this._container = container;

            //Unpack container
            UnpackContainer();

            int seed = 50;

            //Build Scenario
            switch (scenarioCode)
            {
                case ScenarioCodes.Square:
                    return BuildSquareScenario(seed);

                case ScenarioCodes.Line:
                    return BuildLineScenario(seed);

                default:
                    throw new ApplicationException("Unhandled Scenario Code; " + scenarioCode);
            }

        }

        private void UnpackContainer()
        {
            world = _container.RetrieveInstance<World>();
            worldManager = _container.RetrieveInstance<WorldManager>();
            playerManager = _container.RetrieveInstance<PlayerManager>();

            foodResource = _container.RetrieveInstance<ResourceTypeInfo>(ResourceTypeCodes.Food);
            moneyResource = _container.RetrieveInstance<ResourceTypeInfo>(ResourceTypeCodes.Money);

            //Improvement Types
            cityImprovement = _container.RetrieveInstance<ImprovementTypeInfo>(ImprovementTypeCodes.City);
            farmImprovement = _container.RetrieveInstance<ImprovementTypeInfo>(ImprovementTypeCodes.Farm);

            //Worker Types
            generalWorker = _container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Generalist);
            farmWorker = _container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Farmer);
            soldierWorker = _container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Soldier);
            adminWorker = _container.RetrieveInstance<WorkerTypeInfo>(WorkerTypeCodes.Administrator);


            //Task Types
            trainFarmerTask = _container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.TrainFarmer);
            trainSoldierTask = _container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.TrainSoldier);
            workImprovementTask = _container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.WorkImprovement);
            repairImprovementTask = _container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.RepairImprovement);
            patrolBordersTask = _container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.DefendTerritory);
            attackNode = _container.RetrieveInstance<NodeTaskTypeInfo>(NodeTaskTypeCodes.AttackNode);

            //Player Tasks
            workerConsumptionTask = _container.RetrieveInstance<PlayerTaskTypeInfo>(PlayerTaskTypeCodes.WorkerConsumption);
            workerSpawnTask = _container.RetrieveInstance<PlayerTaskTypeInfo>(PlayerTaskTypeCodes.WorkerSpawn);
        }


        private Dictionary<Node, MapPoint> BuildLineScenario(int seed)
        {
            var rand = new Random(seed);

            InitializePlayers(2);

            var player1 = players[0];
            var player2 = players[1];

            Dictionary<Node, MapPoint> map = new Dictionary<Node, MapPoint>();

            //Create P1Node
            var p1Capital = CreateNode(player1, cityImprovement);

            map.Add(p1Capital, new MapPoint(1, 1));

            //Start with player 1 capital
            Node previous = p1Capital;
            Node current = null;

            //Create line of nodes
            for (int i = 0; i < 10; i++)
            {
                //Create Node and Join
                current = CreateNode(null, farmImprovement);
                CreateJoin(previous, current);

                //Add To Map
                map.Add(current, new MapPoint(i + 2, 1));

                //Switch
                previous = current;
                current = null;
            }

            var p2Capital = CreateNode(player2, cityImprovement);
            map.Add(p2Capital, new MapPoint(12, 1));

            //End with player 2 capital
            CreateJoin(previous, p2Capital);


            //Setup Economy
            player1.Resources[foodResource] = 100;
            player1.Resources[moneyResource] = 100;

            player2.Resources[foodResource] = 50;
            player2.Resources[moneyResource] = 50;

            //Retur Node Map
            return map;
        }

        private Dictionary<Node, MapPoint> BuildSquareScenario(int seed)
        {
            //TODO: Use Random Seed
            var rand = new Random(seed);

            //Create Players
            InitializePlayers(4);
            var player1 = players[0];
            var player2 = players[1];
            var player3 = players[2];
            var player4 = players[3];

            //Setup Resources
            player1.Resources[foodResource] = 200;
            player1.Resources[moneyResource] = 200;

            player2.Resources[foodResource] = 150;
            player2.Resources[moneyResource] = 150;

            player3.Resources[foodResource] = 100;
            player3.Resources[moneyResource] = 100;

            player4.Resources[foodResource] = 50;
            player4.Resources[moneyResource] = 50;

            //Create Capitals
            var p1Capital = CreateNode(player1, cityImprovement);
            var p2Capital = CreateNode(player2, cityImprovement);
            var p3Capital = CreateNode(player3, cityImprovement);
            var p4Capital = CreateNode(player4, cityImprovement);

            //Create Map Set Capitals
            Node[,] mapArray = new Node[5, 5];
            mapArray[0, 0] = p1Capital;
            mapArray[4, 4] = p2Capital;
            mapArray[0, 4] = p3Capital;
            mapArray[4, 0] = p4Capital;

            //Fill Map
            for (int x = 0; x < mapArray.GetLength(0); x++)
            {
                for (int y = 0; y < mapArray.GetLength(1); y++)
                {
                    //Skip pre populated nodes (capital)
                    if (mapArray[x, y] != null)
                        continue;

                    //Create Node and Join
                    mapArray[x, y] = CreateNode(null, farmImprovement);
                }
            }

            //Create Joins
            int[][] VectorMap = new int[][] { new int[] { 1, 0 }, new int[] { 1, 1 }, new int[] { 0, 1 }, new int[] { -1, 1 } };
            for (int x = 0; x < mapArray.GetLength(0); x++)
            {
                for (int y = 0; y < mapArray.GetLength(1); y++)
                {
                    var fromNode = mapArray[x, y];

                    foreach (var vector in VectorMap)
                    {
                        var joinToX = x + vector[0];
                        var joinToY = y + vector[1];

                        if (joinToX < 0 || joinToX > mapArray.GetLength(0) - 1)
                            continue;

                        if (joinToY < 0 || joinToY > mapArray.GetLength(1) - 1)
                            continue;

                        var toNode = mapArray[joinToX, joinToY];
                        CreateJoin(fromNode, toNode);
                    }
                }
            }

            //Generate Map Dictionary
            Dictionary<Node, MapPoint> map = new Dictionary<Node, MapPoint>();
            for (int x = 0; x < mapArray.GetLength(0); x++)
            {
                for (int y = 0; y < mapArray.GetLength(1); y++)
                {
                    map.Add(mapArray[x, y], new MapPoint(x + 1, y + 1));
                }
            }

            return map;
        }


        private void InitializePlayers(int playerCount)
        {
            //Create Players
            for (int i = 0; i < playerCount; i++)
            {
                players.Add(CreatePlayer("P" + i.ToString()));
            }
        }

        private Player CreatePlayer(string playerName)
        {
            //Create Player
            var player = worldManager.AddPlayer(playerName);

            //Add Default tasks, will possibly need to be extended.
            playerManager.AddPlayerTask(player, workerSpawnTask);
            playerManager.AddPlayerTask(player, workerConsumptionTask);

            return player;
        }

        private Node CreateNode(Player owner = null, ImprovementTypeInfo typeInfo = null)
        {
            //Create Node, assign ownership and set improvement
            var node = worldManager.AddNode();

            if (typeInfo != null)
                worldManager.SetNodeImprovement(node, typeInfo);

            if (owner != null)
                worldManager.SetNodeOwner(node, owner);

            return node;
        }

        private void CreateJoin(Node from, Node to)
        {
            worldManager.AddJoin(from, to);
        }
    }
}
