﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Core;

using Buildcraft.Core.NodeTasks.Conditions;
using Buildcraft.Core.NodeTasks.Effects;
using Buildcraft.Core.PlayerTasks;
using Buildcraft.Core.PlayerTasks.Effects;
using Buildcraft.Core.Interfaces;
using Buildcraft.Core.TaskProcessors;
using Buildcraft.Core.Interfaces.UnitTasks;
using Buildcraft.Core.UnitTasks;
using Buildcraft.Core.UnitTasks.WorkImprovement;
using Buildcraft.Core.UnitTasks.AttackNode;
using Buildcraft.Core.UnitTasks.DefendNode;
using Buildcraft.Core.UnitTasks.DefendRegion;

namespace Buildcraft.ScenarioBuilder
{
    public class WorldInitializer
    {
        //TODO: Refactor - Break down parts

        private IBuildcraftContainer _container;

        //World and Tasks
        private World world;

        private WorldManager worldManager;
        private PlayerManager playerManager;
        private ITaskScheduler scheduler;


        //Players
        List<Player> players = new List<Player>();

        //Player Capitals
        List<Node> capitals = new List<Node>();

        //Improvement Types
        protected ImprovementTypeInfo cityImprovement;
        protected ImprovementTypeInfo farmImprovement;

        //Resource Types
        protected ResourceTypeInfo foodResource;
        protected ResourceTypeInfo moneyResource;

        //Worker Types
        protected WorkerTypeInfo generalWorker;
        protected WorkerTypeInfo farmWorker;
        protected WorkerTypeInfo soldierWorker;
        protected WorkerTypeInfo adminWorker;

        //Task Types
        protected NodeTaskTypeInfo trainFarmerTask;
        protected NodeTaskTypeInfo trainSoldierTask;
        protected NodeTaskTypeInfo workImprovementTask;
        protected NodeTaskTypeInfo repairImprovementTask;
        protected NodeTaskTypeInfo patrolBordersTask;
        protected NodeTaskTypeInfo attackNode;
        protected NodeTaskTypeInfo sabotageImprovement;

        //Player Tasks
        protected PlayerTaskTypeInfo workerConsumptionTask;
        protected PlayerTaskTypeInfo workerSpawnTask;

        //Unit Types
        protected UnitTypeInfo generalUnit;
        protected UnitTypeInfo farmUnit;
        protected UnitTypeInfo soldierUnit;
        protected UnitTypeInfo adminUnit;

        //Unit Task Types
        protected IUnitTaskInfo workImprovementUnitTask;
        protected IUnitTaskInfo attackNodeUnitTask;
        protected IUnitTaskInfo defendNodeUnitTask;
        protected IUnitTaskInfo defendRegionUnitTask;


        /// <summary>
        /// 
        /// </summary>
        /// <param name="container"></param>
        /// <param name="scenarioCode"></param>
        /// <returns>A map of nodes to locations.</returns>
        public void InitializeContainer(IBuildcraftContainer container)
        {
            if (container == null)
                throw new ArgumentNullException();

            this._container = container;

            //Unpack container and configure world
            UnpackContainer();
            InitializeWorld();
        }

        private void UnpackContainer()
        {
            world = _container.RetrieveInstance<World>();
            worldManager = _container.RetrieveInstance<WorldManager>();
            playerManager = _container.RetrieveInstance<PlayerManager>();
            scheduler = _container.RetrieveInstance<ITaskScheduler>();
        }


        private void InitializeWorld()
        {
            InitializeResourceTypes();
            InitializeImprovementTypes();
            InitializeWorkerTypes();
            InitializeUnitTypes();
            InitializeNodeConditions();
            InitializeNodeEffects();
            InitializePlayerEffects();
            InitializeNodeTaskTypes();
            InitializeUnitTaskTypes();
            InitializePlayerTaskTypes();
            InitializeProcesses();

            _container.Initialize();
        }


        private void InitializeResourceTypes()
        {
            foodResource = _container.AddResourceTypeInfo(ResourceTypeCodes.Food);
            moneyResource = _container.AddResourceTypeInfo(ResourceTypeCodes.Money);
        }

        private void InitializeNodeConditions()
        {
            _container.RegisterInstance<ICondition>(
               new TrainWorkerCondition()
               {
                   MaxWorkers = 1,
                   MinWorkers = 1,
                   ImprovementRestriction = cityImprovement,
                   TargetWorkerType = farmWorker
               },
               NodeConditionCodes.TrainFarmerCondition);

            _container.RegisterInstance<ICondition>(
               new TrainWorkerCondition()
               {
                   MaxWorkers = 1,
                   MinWorkers = 1,
                   ImprovementRestriction = cityImprovement,
                   TargetWorkerType = soldierWorker
               },
               NodeConditionCodes.TrainSoldierCondition);

            _container.RegisterInstance<ICondition>(
                new RepairImprovementCondition()
                {
                    MinWorkers = 1,
                    MaxWorkers = 5
                },
                NodeConditionCodes.RepairImprovementCondition);

            _container.RegisterInstance<ICondition>(
                new PatrolBorderCondition()
                {
                    ImprovementRestriction = cityImprovement,
                    MinWorkers = 1,
                    MaxWorkers = 5
                },
                NodeConditionCodes.DefendTerritoryCondition);

            _container.RegisterInstance<ICondition>(
                 new AttackNodeCondition()
                 {
                     MinWorkers = 1,
                     MaxWorkers = 5
                 },
                 NodeConditionCodes.AttackNodeCondition);

            _container.RegisterInstance<ICondition>(
                new SabotageImprovementCondition()
                {
                    MinWorkers = 1,
                    MaxWorkers = 5
                },
                NodeConditionCodes.SabotageImprovementCondition);


            _container.RegisterInstance<ICondition>(
                new WorkImprovementCondition(),
                NodeConditionCodes.WorkImprovementCondition);
        }

        private void InitializeNodeEffects()
        {
            _container.RegisterInstance<IEffect<NodeTask>>(
                new ChangeWorkerTypeEffect(WorkerTypeCodes.Farmer),
                NodeEffectCodes.TrainFarmerEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new ChangeWorkerTypeEffect(WorkerTypeCodes.Soldier),
                NodeEffectCodes.TrainSoldierEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new RepairImprovementEffect(),
                NodeEffectCodes.RepairImprovementEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new DefendTerritoryEffect()
                {
                    DefencePerWorker = 2,
                    SpecialistType = soldierWorker,
                    SpecialistMultiplier = 2
                },
                NodeEffectCodes.DefendTerritoryEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                new AttackNodeEffect()
                {
                    AttackPerWorker = 1,
                    SpecialistMultiplier = 2,
                    SpecialistType = soldierWorker
                },
                NodeEffectCodes.AttackNodeEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                  new SabotageImprovementEffect(),
                  NodeEffectCodes.SabotageImprovementEffect);

            _container.RegisterInstance<IEffect<NodeTask>>(
                    new WorkImprovementEffect(),
                    NodeEffectCodes.WorkImprovementEffect);
        }

        private void InitializePlayerEffects()
        {
            var workerConsumptionEffect = new WorkerConsumptionEffect();
            _container.RegisterInstance<IEffect<PlayerTask>>(workerConsumptionEffect, PlayerEffectCodes.WorkerConsumptionEffects);

            var workerSpawnEffect = new WorkerSpawnEffect();
            workerSpawnEffect.SpawnType = generalWorker;
            _container.RegisterInstance<IEffect<PlayerTask>>(workerSpawnEffect, PlayerEffectCodes.WorkerSpawnEffect);
        }

        private void InitializeImprovementTypes()
        {
            cityImprovement = _container.AddImprovementTypeInfo(ImprovementTypeCodes.City, 1, 5);
            cityImprovement.ResourceProductionType = moneyResource;
            cityImprovement.ResourceProductionRate = 20;

            farmImprovement = _container.AddImprovementTypeInfo(ImprovementTypeCodes.Farm, 1, 5);
            farmImprovement.SpecialistWorkerType = farmWorker;
            farmImprovement.ResourceProductionType = foodResource;
            farmImprovement.ResourceProductionRate = 20;
        }

        private void InitializeWorkerTypes()
        {
            //Worker Types
            generalWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Generalist, 0);
            generalWorker.Consumption[foodResource] = 2;
            generalWorker.Consumption[moneyResource] = 1;

            farmWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Farmer, 2);
            farmWorker.Consumption[foodResource] = 2;
            farmWorker.Consumption[moneyResource] = 2;

            soldierWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Soldier, 3);
            soldierWorker.Consumption[foodResource] = 2;
            soldierWorker.Consumption[moneyResource] = 4;

            adminWorker = _container.AddWorkerTypeInfo(WorkerTypeCodes.Administrator, 1);
            adminWorker.Consumption[moneyResource] = 1;
            adminWorker.Consumption[foodResource] = 2;
        }

        private void InitializeUnitTypes()
        {
            //Unit Types
            generalUnit = _container.AddUnitTypeInfo(WorkerTypeCodes.Generalist);
            generalUnit.Consumption[foodResource] = 2;
            generalUnit.Consumption[moneyResource] = 1;

            farmUnit = _container.AddUnitTypeInfo(WorkerTypeCodes.Farmer);
            farmUnit.Consumption[foodResource] = 2;
            farmUnit.Consumption[moneyResource] = 2;

            soldierUnit = _container.AddUnitTypeInfo(WorkerTypeCodes.Soldier);
            soldierUnit.Consumption[foodResource] = 2;
            soldierUnit.Consumption[moneyResource] = 4;

            adminUnit = _container.AddUnitTypeInfo(WorkerTypeCodes.Administrator);
            adminUnit.Consumption[moneyResource] = 1;
            adminUnit.Consumption[foodResource] = 2;
        }

        private void InitializeNodeTaskTypes()
        {

            trainSoldierTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.TrainSoldier,
                NodeConditionCodes.TrainSoldierCondition,
                NodeEffectCodes.TrainSoldierEffect,
                100,
                false);

            trainFarmerTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.TrainFarmer,
                NodeConditionCodes.TrainFarmerCondition,
                NodeEffectCodes.TrainFarmerEffect,
                100,
                false);

            repairImprovementTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.RepairImprovement,
                NodeConditionCodes.RepairImprovementCondition,
                NodeEffectCodes.RepairImprovementEffect,
                100,
                false);

            patrolBordersTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.DefendTerritory,
                NodeConditionCodes.DefendTerritoryCondition,
                NodeEffectCodes.DefendTerritoryEffect,
                100,
                true);

            attackNode = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.AttackNode,
                NodeConditionCodes.AttackNodeCondition,
                NodeEffectCodes.AttackNodeEffect,
                100,
                false);

            sabotageImprovement = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.SabotageImprovement,
                NodeConditionCodes.SabotageImprovementCondition,
                NodeEffectCodes.SabotageImprovementEffect,
                100,
                false);

            workImprovementTask = _container.AddNodeTaskTypeInfo(
                NodeTaskTypeCodes.WorkImprovement,
                NodeConditionCodes.WorkImprovementCondition,
                NodeEffectCodes.WorkImprovementEffect,
                100,
                true);
        }

        private void InitializeUnitTaskTypes()
        {
            this.workImprovementUnitTask = new WorkImprovementTaskTypeInfo();
            _container.RegisterInstance<IUnitTaskInfo>(workImprovementUnitTask, workImprovementUnitTask.TaskCode);

            this.attackNodeUnitTask = new AttackNodeTaskTypeInfo();
            _container.RegisterInstance<IUnitTaskInfo>(attackNodeUnitTask, attackNodeUnitTask.TaskCode);

            this.defendNodeUnitTask = new DefendNodeTaskTypeInfo();
            _container.RegisterInstance<IUnitTaskInfo>(defendNodeUnitTask, defendNodeUnitTask.TaskCode);

            this.defendRegionUnitTask = new DefendRegionTaskTypeInfo();
            _container.RegisterInstance<IUnitTaskInfo>(defendRegionUnitTask, defendRegionUnitTask.TaskCode);
        }

        private void InitializePlayerTaskTypes()
        {
            workerConsumptionTask = _container.AddPlayerTaskTypeInfo(
                PlayerTaskTypeCodes.WorkerConsumption,
                PlayerEffectCodes.WorkerConsumptionEffects,
                150,
                true);

            workerSpawnTask = _container.AddPlayerTaskTypeInfo(
                PlayerTaskTypeCodes.WorkerSpawn,
                PlayerEffectCodes.WorkerSpawnEffect,
                200,
                true);
        }

        private void InitializeProcesses()
        {
            scheduler.RegisterTaskProcessor(new CombatProcessor());
            scheduler.RegisterTaskProcessor(new TaskProductionProcessor());
            scheduler.RegisterTaskProcessor(new TaskRunningCostProcessor());
        }
    }

}
