﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Buildcraft.Core;
using Buildcraft.Core.Model;
using Buildcraft.Core.Components;
using Buildcraft.Controls;
using Buildcraft.AI;
using Buildcraft.Core.Interfaces;
using Buildcraft.ScenarioBuilder;


namespace Buildcraft
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, IHostNotify
    {
        private IBuildcraftContainer _container;
        private TickWorldAction _tickWorldAction;
        private World _world;

        Dictionary<Node, NodeDetails> _uiNodeDictionary = new Dictionary<Node, NodeDetails>();
        List<IRebindable> _rebindableElements = new List<IRebindable>();

        private List<AIPlayer> _aiPlayers = new List<AIPlayer>();
        private Dictionary<Node, MapPoint> _map;

        public MainWindow()
        {
            InitializeComponent();
        }


        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            CreateNewGame();
            BeginGameLoop();
        }


        public void CreateNewGame()
        {
            //Create, load and initialize container
            _container = new TestContainer();
            ContainerUtilities.LoadContainer(_container);

            //Load IHostNotify Implementations
            _container.RegisterInstance<IHostNotify>(this);

            //Initialize Container
            _container.Initialize();

            //Pull dependencies from container
            _tickWorldAction = _container.RetrieveInstance<TickWorldAction>();
            _world = _container.RetrieveInstance<World>();

            //Setup World
            WorldInitializer builder = new WorldInitializer();
            builder.InitializeContainer(_container);
            ScenarioAssembler assembler = new ScenarioAssembler();
            _map = assembler.BuildScenario(_container, ScenarioCodes.Square);

            //Load Players
            foreach(var player in _world.Players)
            {
                var details = new PlayerDetails();
                details.Player = player;

                //Add details list
                _rebindableElements.Add(details);
                _aiPlayers.Add(new AIPlayer(details.Player, _container));

                playerDetailStack.Children.Add(details);
            }

            //Load Map
            LoadScenario(_map);
        }

        private void LoadScenario(Dictionary<Node, MapPoint> _map)
        {
            int regionSize = 40;
            int unitMultiplier = 50;

            //Clear old dictionaries
            _uiNodeDictionary.Clear();

            //Draw Joins (Do this first so the nodes are rendered on top of the lines)
            SolidColorBrush brush = new SolidColorBrush(Colors.Black);
            foreach (var fromKvp in _map)
            {
                //Unpack
                var fromNode = fromKvp.Key;
                var fromPosition = fromKvp.Value;

                foreach (var join in fromNode.Joins)
                {
                    var toNode = join.ToNode;
                    var toPosition = _map[toNode];

                    var line = new Line();
                    line.Stroke = brush;
                    line.StrokeThickness = 2;

                    line.X1 = (fromPosition.X * unitMultiplier) + (regionSize / 2);
                    line.X2 = (toPosition.X * unitMultiplier) + (regionSize / 2);

                    line.Y1 = (fromPosition.Y * unitMultiplier) + (regionSize / 2);
                    line.Y2 = (toPosition.Y * unitMultiplier) + (regionSize / 2);

                    worldCanvas.Children.Add(line);
                }
            }

            //Draw Nodes
            foreach (var kvp in _map)
            {
                //Unpack
                var node = kvp.Key;
                var position = kvp.Value;

                var uiNode = new NodeDetails();
                uiNode.GameNode = node;
                uiNode.Width = regionSize;
                uiNode.Height = regionSize;
                uiNode.RenderTransform = new TranslateTransform(position.X * unitMultiplier, position.Y * unitMultiplier);

                //Add to enviroment
                worldCanvas.Children.Add(uiNode);
                _rebindableElements.Add(uiNode);
                _uiNodeDictionary.Add(node, uiNode);
            }
        }

        private void BeginGameLoop()
        {
            CompositionTarget.Rendering += new EventHandler(CompositionTarget_Rendering);
        }

        /// <summary>
        /// Frame Event handler
        /// </summary>
        private void CompositionTarget_Rendering(object sender, EventArgs e)
        {
            int frameSize = 2;

            //Tick World
            _tickWorldAction.TickWorld(frameSize);

            foreach (var aiPlayer in _aiPlayers)
            {
                aiPlayer.Tick(frameSize);
            }

            //Refresh
            foreach (var item in _rebindableElements)
            {
                item.Rebind();
            }
        }

        #region IHostNotify Members

        public void TaskAdded(NodeTask task)
        {
            _uiNodeDictionary[task.Node].AddNodeTask(task);
            //OutputText(string.Format("Task Added: {0}", task.ToString()));
        }

        public void TaskRemoved(NodeTask task)
        {
            _uiNodeDictionary[task.Node].RemoveNodeTask(task);
            //OutputText(string.Format("Task Removed: {0}", task.ToString()));
        }

        public void ResourcesProduced(NodeTask task, ResourceTypeInfo resourceType, int production)
        {
            //TODO: Play Animation on Node

            if (production > 0)
                OutputText(string.Format("{0} production - {1} : {2}", task.Owner.Name, resourceType, production));
        }

        public void GameOver(Player winner)
        {
            OutputText(string.Format("GAME OVER! Player {0} wins.", winner.Name));
        }

        #endregion

        private void OutputText(string text)
        {
            this.consoleOutput.AppendText(text + Environment.NewLine);
            this.consoleOutput.ScrollToEnd();
        }
    }

}
